from typing import Union, Optional
import re
import os
import pandas as pd
import polars as pl
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
import statsmodels as sm
import statsmodels.api as sma
import statsmodels.formula.api as smf
from statsmodels.stats.outliers_influence import variance_inflation_factor
from math import ceil
from scipy import stats
from sklearn import metrics
from sklearn.model_selection import StratifiedShuffleSplit, GridSearchCV, StratifiedKFold, KFold
from scipy.special import expit
from pyrsm.utils import ifelse, expand_grid, check_dataframe
from pyrsm.stats import weighted_mean, weighted_sd
from pyrsm.notebook import load_state, save_state
from .perf import auc


def make_train(
    data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
    strat_var: str | list[str] = None,
    test_size: float = 0.2,
    random_state: int = 1234,
):
    """
    Use stratified sampling on one or more variables to create a training
    variable for your dataset
    """
    splits = StratifiedShuffleSplit(
        n_splits=1, test_size=test_size, random_state=random_state
    )
    data = check_dataframe(data)
    training_var = np.zeros(data.shape[0])
    for train_index, test_index in splits.split(data, data[strat_var]):
        training_var[train_index] = 1
    return training_var


def extract_evars(model, cn):
    """
    Extract a list of the names of the explanatory variables in a statsmodels model
    """
    pattern = r"\b\w+\b"
    evars = re.findall(pattern, model.formula)[1:]
    evars = [v for v in evars if v in cn]
    return [v for i, v in enumerate(evars) if v not in evars[:i]]


def extract_rvar(model, cn):
    """
    Extract name of the response variable in a statsmodels model
    """
    pattern = r"\b\w+\b"
    return re.findall(pattern, model.formula)[0]


def convert_to_list(v):
    if v is None:
        return []
    elif isinstance(v, list):
        return v
    elif isinstance(v, str):
        return [v]
    else:
        return list(v)


def convert_binary(data, rvar, lev):
    cb = ifelse(data[rvar] == lev, 1, ifelse(data[rvar].isna(), np.nan, 0))
    data = data.drop(columns=rvar).copy()
    data.loc[:, rvar] = cb
    return data


def is_binary(series):
    """
    Efficiently check if a series has more than 2 unique values.
    Stops checking as soon as it finds a third unique value.
    """
    seen = set()
    series = series.dropna()
    for val in series:
        seen.add(val)
        if len(seen) > 2:
            return True
    return False


def check_binary(data, rvar, lev):
    unique_levels = data[rvar].unique()
    data = convert_binary(data, rvar, lev)
    rvar_sum = data[rvar].sum()
    if rvar_sum == 0 or rvar_sum == data.shape[0]:
        raise ValueError(
            f"All converted response values are {1 if rvar_sum == data.shape[0] else 0}. "
            f"Available levels in {rvar}: {unique_levels} and '{lev}' was selected."
        )
    else:
        return data


def conditional_get_dummies(df, drop_nonvarying=True):
    for column in df.select_dtypes(include=["object", "category"]).columns:
        if not is_binary(df[column]):
            dummies = pd.get_dummies(df[column], prefix=column, drop_first=True)
        else:
            dummies = pd.get_dummies(df[column], prefix=column)
        if not drop_nonvarying and len(dummies.columns) == 0:
            dummies = pd.get_dummies(df[column], prefix=column, drop_first=False)

        df = pd.concat([df.drop(column, axis=1), dummies], axis=1)
    return df


def get_dummies(df, drop_first=True, drop_nonvarying=True):
    for column in df.select_dtypes(include=["object", "category"]).columns:
        dummies = pd.get_dummies(df[column], prefix=column, drop_first=drop_first)
        if not drop_nonvarying and len(dummies.columns) == 0:
            dummies = pd.get_dummies(df[column], prefix=column, drop_first=False)
        df = pd.concat([df.drop(column, axis=1), dummies], axis=1)
    return df


def sig_stars(pval):
    pval = np.nan_to_num(pval, nan=1.0)
    cutpoints = np.array([0.001, 0.01, 0.05, 0.1, np.inf])
    symbols = np.array(["***", "**", "*", ".", " "])
    return [symbols[p < cutpoints][0] for p in pval]


def get_mfit(fitted) -> tuple[Optional[dict], Optional[str]]:
    """
    Get model fit statistics

    Parameters
    ----------
    fitted : A fitted linear or logistic 4egression model
    """

    mfit_dct = None
    model_type = None
    if isinstance(fitted, sm.genmod.generalized_linear_model.GLMResultsWrapper):
        fw = None
        if fitted.model._has_freq_weights:
            fw = fitted.model.freq_weights

        # gets same results as in R
        lrtest = -2 * (fitted.llnull - fitted.llf)
        mfit_dct = {
            "pseudo_rsq_mcf": [1 - fitted.llf / fitted.llnull],
            "pseudo_rsq_mcf_adj": [1 - (fitted.llf - fitted.df_model) / fitted.llnull],
            "AUC": [auc(fitted.model.endog, fitted.fittedvalues, weights=fw)],
            "log_likelihood": fitted.llf,
            "AIC": [fitted.aic],
            "BIC": [fitted.bic_llf],
            # "chisq": [fitted.pearson_chi2],
            "chisq": [lrtest],
            "chisq_df": [fitted.df_model],
            # "chisq_pval": [1 - stats.chi2.cdf(fitted.pearson_chi2, fitted.df_model)],
            "chisq_pval": [stats.chi2.sf(lrtest, fitted.df_model)],
            "nobs": [fitted.nobs],
        }

        model_type = "logistic"

    elif isinstance(fitted, sm.regression.linear_model.RegressionResultsWrapper):
        mfit_dct = {
            "rsq": [fitted.rsquared],
            "rsq_adj": [fitted.rsquared_adj],
            "fvalue": [fitted.fvalue],
            "ftest_df_model": [fitted.df_model],
            "ftest_df_resid": [fitted.df_resid],
            "ftest_pval": [fitted.f_pvalue],
            "nobs": [fitted.nobs],
        }

        model_type = "regression"

    return mfit_dct, model_type


def or_ci(fitted, alpha=0.05, intercept=False, importance=False, data=None, dec=3):
    """
    Confidence interval for Odds ratios

    Parameters
    ----------
    fitted : A fitted logistic regression model
    alpha : float
        Significance level
    intercept : bool
        Include intercept in output (True or False)
    importance : int
        Calculate variable importance. Only meaningful if data
        used in estimation was standardized prior to model
        estimation
    data : Pandas dataframe
        Unstandardized data used to calculate descriptive
        statistics
    dec : int
        Number of decimal places to use in rounding

    Returns
    -------
    Pandas dataframe with Odd-ratios and confidence intervals
    """

    df = pd.DataFrame(np.exp(fitted.params), columns=["OR"]).dropna()
    df["OR%"] = 100 * ifelse(df["OR"] < 1, -(1 - df["OR"]), df["OR"] - 1)

    low, high = [100 * alpha / 2, 100 * (1 - (alpha / 2))]
    df[[f"{low}%", f"{high}%"]] = np.exp(fitted.conf_int(alpha=alpha))
    df["p.values"] = ifelse(fitted.pvalues < 0.001, "< .001", fitted.pvalues.round(dec))
    df["  "] = sig_stars(fitted.pvalues)
    df["OR%"] = [f"{round(o, max(dec-2, 0))}%" for o in df["OR%"]]
    df = df.reset_index()

    if importance:
        df["dummy"] = df["index"].str.contains("[T", regex=False)
        df["importance"] = (
            pd.DataFrame().assign(OR=df["OR"], ORinv=1 / df["OR"]).max(axis=1)
        )

    if isinstance(data, pd.DataFrame):
        # using a fake response variable variable
        data = data.assign(__rvar__=1).copy()
        form = "__rvar__ ~ " + fitted.model.formula.split("~", 1)[1]
        exog = pd.DataFrame(smf.logit(formula=form, data=data).exog)
        weights = fitted._freq_weights
        if sum(weights) > len(weights):

            def wmean(x):
                return weighted_mean(x, weights)

            def wstd(x):
                return weighted_sd(pd.DataFrame(x), weights)[0]

            df = pd.concat(
                [df, exog.apply([wmean, wstd, "min", "max"]).T],
                axis=1,
            )
        else:
            df = pd.concat([df, exog.apply(["mean", "std", "min", "max"]).T], axis=1)

    if intercept is False:
        df = df.loc[df["index"] != "Intercept"]

    return df.round(dec)


def or_plot(fitted, alpha=0.05, intercept=False, incl=None, excl=None, figsize=None):
    """
    Odds ratio plot

    Parameters
    ----------
    fitted : A fitted logistic regression model
    alpha : float
        Significance level
    intercept : bool
        Include intercept in odds-ratio plot (True or False)
    incl : str or list of strings
        Variables to include in the odds-ratio plot. All will be included by default
    excl : str or list of strings
        Variables to exclude from the odds-ratio plot. None are excluded by default

    Returns
    -------
    Matplotlit object
        Plot of Odds ratios
    """

    # iloc to reverse order
    df = or_ci(fitted, alpha=alpha, intercept=intercept, dec=100).dropna().iloc[::-1]

    if incl is not None:
        incl = ifelse(isinstance(incl, str), [incl], incl)
        rx = "(" + "|".join([f"^{v}$|^{v}\\[" for v in incl]) + ")"
        incl = df["index"].str.match(rf"{rx}")
        if intercept:
            incl[0] = True
        df = df[incl]

    if excl is not None and excl != []:
        excl = ifelse(isinstance(excl, str), [excl], excl)
        rx = "(" + "|".join([f"^{v}$|^{v}\\[" for v in excl]) + ")"
        excl = df["index"].str.match(rf"{rx}")
        if intercept:
            excl[0] = False
        df = df[~excl]

    low, high = [100 * alpha / 2, 100 * (1 - (alpha / 2))]
    err = [df["OR"] - df[f"{low}%"], df[f"{high}%"] - df["OR"]]

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot()
    ax.axvline(1, ls="dashdot")
    ax.errorbar(x="OR", y="index", data=df, xerr=err, fmt="none")
    ax.scatter(x="OR", y="index", data=df)
    ax.set_xscale("log")
    ax.xaxis.set_minor_formatter(ticker.NullFormatter())
    ax.xaxis.set_major_locator(ticker.LogLocator(subs=[0.1, 0.2, 0.5, 1, 2, 5, 10]))
    ax.xaxis.set_major_formatter(ticker.StrMethodFormatter("{x:.1f}"))
    ax.set(xlabel="Odds-ratio")
    return ax


def vif(fitted, dec=3):
    """
    Calculate the Variance Inflation Factor (VIF) associated with each
    exogenous variable

    Status
    ------
    WIP port of VIF calculation from R's car:::vif.default to Python

    Parameters
    ----------
    fitted : A fitted (logistic) regression model
    dec : int
        Number of decimal places to use in rounding

    Returns
    -------
    Pandas dataframe sorted by VIF score
    """

    if hasattr(fitted, "model"):
        model = fitted.model
    else:
        # legacy for when only an un-fitted model was accepted
        model = fitted

    is_categorical = {
        item.split("[T.")[0]: ("[T." in item) for item in model.exog_names
    }
    if sum(is_categorical.values()) > 0:
        evar = list(is_categorical.keys())[1:]
        exog = model.exog[:, 1:]
        Xcorr = np.linalg.det(np.corrcoef(exog, rowvar=False))
        df = pd.DataFrame(exog, columns=model.exog_names[1:])
        vif = []
        for col, cat in is_categorical.items():
            if col == "Intercept":
                continue
            elif cat:
                select = [f"{col}[T." in c for c in df.columns]
                Vcorr = np.linalg.det(df.loc[:, select].corr().values)
                drop = [f"{col}[T." not in c for c in df.columns]
                Ocorr = np.linalg.det(df.loc[:, drop].corr().values)
            else:
                Vcorr = 1
                Ocorr = np.linalg.det(df.drop(col, axis=1).corr().values)

            vif.append(Vcorr * Ocorr / Xcorr)
        df = pd.DataFrame(evar, columns=["index"])
    else:
        vif = [
            variance_inflation_factor(model.exog, i) for i in range(model.exog.shape[1])
        ]
        df = pd.DataFrame(model.exog_names, columns=["index"])

    df["vif"] = vif
    df["Rsq"] = 1 - 1 / df["vif"]

    if "Intercept" in model.exog_names:
        df = df[df["index"] != "Intercept"]

    df = df.sort_values("vif", ascending=False).set_index("index")
    df.index.name = None

    if dec is not None:
        df = df.round(dec)

    return df


def predict_ci(fitted, df, conf=0.95, alpha=None):
    """
    Compute predicted probabilities with confidence intervals based on a
    logistic regression model

    Parameters
    ----------
    fitted : Logistic regression model fitted using the statsmodels formula interface
    df : Pandas dataframe with input data for prediction
    conf : float
        Confidence level (0-1). Default is 0.95
    alpha : float
        Significance level (0-1). Default is 0.05

    Returns
    -------
    Pandas DataFrame with probability predictions and lower and upper confidence bounds

    Example
    -------
    import numpy as np
    import statsmodels.formula.api as smf
    import pandas as pd

    # simulate data
    np.random.seed(1)
    x1 = np.arange(100)
    x2 = pd.Series(["a", "b", "c", "a"], dtype="category").sample(100, replace=True)
    y = (x1 * 0.5 + np.random.normal(size=100, scale=10) > 30).astype(int)
    df = pd.DataFrame({"y": y, "x1": x1, "x2": x2})

    # estimate the model
    model = smf.logit(formula="y ~ x1 + x2", data=df).fit()
    model.summary()
    pred = predict_ci(model, df)

    plt.clf()
    plt.plot(x1, pred["prediction"])
    plt.plot(x1, pred["2.5%"], color='black', linestyle="--", linewidth=0.5)
    plt.plot(x1, pred["97.5%"], color='black', linestyle="--", linewidth=0.5)
    plt.show()
    """

    if conf < 0 or conf > 1:
        raise ValueError(
            "Confidence level (conf) must be a numeric value between 0 and 1"
        )

    if alpha is not None:
        raise ValueError(
            "The alpha argument has been deprecated. Use the confidence level (conf) instead (1-alpha)."
        )

    # generate predictions
    prediction = fitted.predict(df)

    # set up the data in df in the same was as the exog data
    # that is part of fitted.model.exog
    # use a fake response variable
    df = df.assign(__rvar__=1).copy()
    form = "__rvar__ ~ " + fitted.model.formula.split("~", 1)[1]
    exog = smf.logit(formula=form, data=df).exog

    low, high = [(1.0 - conf) / 2.0, (1.0 - (1.0 - conf) / 2.0)]
    Xb = np.dot(exog, fitted.params)
    se = np.sqrt((exog.dot(fitted.cov_params()) * exog).sum(-1))
    me = stats.norm.ppf(high) * se

    if isinstance(fitted, sm.genmod.generalized_linear_model.GLMResultsWrapper):
        return pd.DataFrame(
            {
                "prediction": prediction,
                f"{low*100:.2f}%": expit(Xb - me),
                f"{high*100:.2f}%": expit(Xb + me),
            }
        )
    elif isinstance(fitted, sm.regression.linear_model.RegressionResultsWrapper):
        return pd.DataFrame(
            {
                "prediction": prediction,
                f"{low*100:.2f}%": Xb - me,
                f"{high*100:.2f}%": Xb + me,
            }
        )


def nobs_dropped(obj):
    if hasattr(obj, "nobs_dropped") and obj.nobs_dropped > 0:
        return f" ({obj.nobs_dropped:,.0f} obs. dropped)"
    else:
        return ""


def model_fit(fitted, dec: int = 3, prn: bool = True) -> Union[str, pd.DataFrame]:
    """
    Compute various model fit statistics for a fitted linear or logistic regression model

    Parameters
    ----------
    fitted : statmodels ols or glm object
        Regression model fitted using statsmodels
    dec : int
        Number of decimal places to use in rounding
    prn : bool
        If True, print output, else return a Pandas dataframe with the results

    Returns
    -------
        If prn is True, print output, else return a Pandas dataframe with the results
    """
    if hasattr(fitted, "df_resid"):
        weighted_nobs = (
            fitted.df_resid + fitted.df_model + int(hasattr(fitted.params, "Intercept"))
        )
        if weighted_nobs > fitted.nobs:
            fitted.nobs = weighted_nobs
    mfit_dct, model_type = get_mfit(fitted)
    if not mfit_dct:
        return "Only linear and logistic regression models are currently supported."

    mfit = pd.DataFrame(mfit_dct)

    if prn:
        if model_type == "logistic":
            output = f"""Pseudo R-squared (McFadden): {mfit.pseudo_rsq_mcf.values[0].round(dec)}
Pseudo R-squared (McFadden adjusted): {mfit.pseudo_rsq_mcf_adj.values[0].round(dec)}
Area under the RO Curve (AUC): {mfit.AUC.values[0].round(dec)}
Log-likelihood: {mfit.log_likelihood.values[0].round(dec)}, AIC: {mfit.AIC.values[0].round(dec)}, BIC: {mfit.BIC.values[0].round(dec)}
Chi-squared: {mfit.chisq.values[0].round(dec)}, df({mfit.chisq_df.values[0]}), p.value {np.where(mfit.chisq_pval.values[0] < .001, "< 0.001", mfit.chisq_pval.values[0].round(dec))} 
Nr obs: {mfit.nobs.values[0]:,.0f}{nobs_dropped(fitted)}"""
        elif model_type == "regression":
            output = f"""R-squared: {mfit.rsq.values[0].round(dec)}, Adjusted R-squared: {mfit.rsq_adj.values[0].round(dec)}
F-statistic: {mfit.fvalue[0].round(dec)} df({mfit.ftest_df_model.values[0]:.0f}, {mfit.ftest_df_resid.values[0]:.0f}), p.value {np.where(mfit.ftest_pval.values[0] < .001, "< 0.001", mfit.ftest_pval.values[0].round(dec))}
Nr obs: {mfit.nobs.values[0]:,.0f}{nobs_dropped(fitted)}"""
        else:
            output = "Model type not supported"
        return output
    else:
        return mfit


def coef_plot(
    fitted,
    alpha: float = 0.05,
    intercept: bool = False,
    incl: str = None,
    excl: list = [],
    figsize: tuple = None,
):
    """
    Coefficient plot

    Parameters
    ----------
    fitted : A fitted linear regression model
    alpha : float
        Significance level
    intercept : bool
        Include intercept in coefficient plot (True or False)
    incl : str or list of strings
        Variables to include in the coefficient plot. All will be included by default
    excl : str or list of strings
        Variables to exclude from the coefficient plot. None are excluded by default

    Returns
    -------
    Matplotlib object
        Plot of Odds ratios
    """
    df = fitted.conf_int(alpha=alpha).reset_index().iloc[::-1]
    df["coefficient"] = fitted.params[df["index"]].dropna().values

    if not intercept:
        df = df.query('index != "Intercept"')

    if incl is not None:
        incl = ifelse(isinstance(incl, str), [incl], incl)
        rx = "(" + "|".join([f"^\b{v}|^{v}\\[" for v in incl]) + ")"
        incl = df["index"].str.match(rf"{rx}")
        if intercept:
            incl[0] = True
        df = df[incl]

    if len(excl) > 0 and excl is not None:
        excl = ifelse(isinstance(excl, str), [excl], excl)
        rx = "(" + "|".join([f"^\b{v}|^{v}\\[" for v in excl]) + ")"
        excl = df["index"].str.match(rf"{rx}")
        if intercept:
            excl[0] = False
        df = df[~excl]

    low, high = [100 * alpha / 2, 100 * (1 - (alpha / 2))]
    df.columns = ["index", f"{low}%", f"{high}%", "coefficient"]
    err = [df["coefficient"] - df[f"{low}%"], df[f"{high}%"] - df["coefficient"]]

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot()
    ax.axvline(0, ls="dashdot")
    ax.errorbar(x="coefficient", y="index", data=df, xerr=err, fmt="none")
    ax.scatter(x="coefficient", y="index", data=df)
    ax.set(xlabel="Coefficient")
    return ax


def coef_ci(fitted, alpha: float = 0.05, intercept: bool = True, dec: int = 3):
    """
    Confidence interval for coefficient from linear regression

    Parameters
    ----------
    fitted : A fitted linear regression model
    alpha : float
        Significance level
    intercept : bool
        Include intercept in the output (True or False)
    dec : int
        Number of decimal places to use in rounding

    Returns
    -------
    Pandas dataframe with regression coefficients and confidence intervals
    """

    df = pd.DataFrame({"coefficient": fitted.params})

    low, high = [100 * alpha / 2, 100 * (1 - (alpha / 2))]
    df[[f"{low}%", f"{high}%"]] = fitted.conf_int(alpha=alpha)

    if dec is None:
        df["p.values"] = ifelse(fitted.pvalues < 0.001, "< .001", fitted.pvalues)
    else:
        df = df.round(dec)
        df["p.values"] = ifelse(
            fitted.pvalues < 0.001, "< .001", fitted.pvalues.round(dec)
        )

    df["  "] = sig_stars(fitted.pvalues)
    if intercept is False:
        df = df[df.index != "Intercept"]

    return df


def evalreg(df, rvar: str, pred: str, dec: int = 3):
    """
    Evaluate regression models. Calculates R-squared, MSE, and MAE

    Parameters
    ----------
    df : Pandas DataFrame or a dictionary of DataFrames with keys to show results for
        multiple model predictions and datasets (training and test)
    rvar : str
        Name of the response variable column in df
    pred : str
        Name of the column, of list of column names, in df with model predictions
    dec : int
        Number of decimal places to use in rounding

    Examples
    --------
    """

    dct = ifelse(isinstance(df, dict), df, {"All": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)

    def calculate_metrics(key, dfm, pm):
        return pd.DataFrame().assign(
            Type=[key],
            predictor=[pm],
            n=[dfm.shape[0]],
            r2=[metrics.r2_score(dfm[rvar], dfm[pm])],
            mse=[metrics.mean_squared_error(dfm[rvar], dfm[pm])],
            mae=[metrics.mean_absolute_error(dfm[rvar], dfm[pm])],
        )

    result = pd.concat(
        [calculate_metrics(key, val, p) for key, val in dct.items() for p in pred],
        axis=0,
    )
    result.index = range(result.shape[0])
    return result.round(dec)


def reg_dashboard(fitted, nobs: int = 1000):
    """
    Plot regression residual dashboard

    Parameters
    ----------
    fitted : Object with fitted values and residuals
    nobs: int
        Number of observations to use for plots.
        Set to None or -1 to plot all values.
        The Residuals vs Order plot will only be valid
        if all observations are plotted
    """
    fig, axes = plt.subplots(3, 2, figsize=(10, 10))
    plt.subplots_adjust(wspace=0.25, hspace=0.4)

    data = pd.DataFrame().assign(
        fitted=fitted.fittedvalues,
        actual=fitted.model.endog,
        resid=fitted.resid,
        std_resid=fitted.resid / np.std(fitted.resid),
        order=np.arange(fitted.model.endog.shape[0]),
    )

    if (nobs != -1 and nobs is not None) and (nobs < data.shape[0]):
        data = data.sample(nobs)

    sns.regplot(x="fitted", y="actual", data=data, ax=axes[0, 0]).set(
        title="Actual vs Fitted values", xlabel="Fitted values", ylabel="Actual values"
    )
    sns.regplot(x="fitted", y="resid", data=data, ax=axes[0, 1]).set(
        title="Residuals vs Fitted values", xlabel="Fitted values", ylabel="Residuals"
    )
    sns.lineplot(x="order", y="resid", data=data, ax=axes[1, 0]).set(
        title="Residuals vs Row order", ylabel="Residuals", xlabel=None
    )
    sma.qqplot(data.resid, line="s", ax=axes[1, 1])
    axes[1, 1].title.set_text("Normal Q-Q plot")
    pdp = data.resid.plot.hist(
        ax=axes[2, 0],
        title="Histogram of residuals",
        xlabel="Residuals",
        rot=0,
        color="slateblue",
    )
    pdp.set_xlabel("Residuals")
    sns.kdeplot(
        data.std_resid, color="green", fill=True, ax=axes[2, 1], common_norm=True
    )

    # from https://stackoverflow.com/a/52925509/1974918
    norm_x = np.arange(-3, +3, 0.01)
    norm_y = stats.norm.pdf(norm_x)
    sns.lineplot(x=norm_x, y=norm_y, lw=1, ax=axes[2, 1]).set(
        title="Residuals vs Normal density", xlabel="Residuals"
    )


def sim_prediction(
    data: pd.DataFrame,
    vary: list = [],
    nnv: int = 5,
    minq: float = 0,
    maxq: float = 1,
) -> pd.DataFrame:
    """
    Simulate data for prediction

    Parameters
    ----------
    data : Pandas DataFrame
    vary : List of column names or Dictionary with keys and values to use
    nnv : int
        Number of values to use to simulate the effect of a numeric variable
    minq : float
        Quantile to use for the minimum value of numeric variables
    maxq : float
        Quantile to use for the maximum value of numeric variables

    Returns:
    ----------
    Pandas DataFrame with values to use for estimation
    """

    def fix_value(s):
        if pd.api.types.is_numeric_dtype(s.dtype):
            return s.mean()
        else:
            return s.value_counts().idxmax()

    dct = {c: [fix_value(data[c])] for c in data.columns}
    dt = data.dtypes
    if isinstance(vary, dict):
        # user provided values and ranges
        for key, val in vary.items():
            dct[key] = val
    else:
        # auto derived values and ranges
        vary = ifelse(isinstance(vary, str), [vary], vary)
        for v in vary:
            if pd.api.types.is_numeric_dtype(data[v].dtype):
                nu = data[v].nunique()
                if nu > 2:
                    dct[v] = np.linspace(
                        np.quantile(data[v], minq),
                        np.quantile(data[v], maxq),
                        min([nu, nnv]),
                    )
                else:
                    dct[v] = [data[v].min(), data[v].max()]
            else:
                dct[v] = data[v].unique()

    return expand_grid(dct, dt)


def scatter_plot(
    fitted, df, nobs: int = 1000, figsize: tuple = None, resid=False
) -> None:
    """
    Scatter plot of explanatory and response variables from a fitted regression

    Parameters
    ----------
    fitted : A fitted linear regression model
    df : Pandas DataFrame
        Data frame with explanatory and response variables
    nobs : int
        Number of observations to use for the scatter plots. The default
        value is 1,000. To use all observations in the plots, use nobs=-1
    figsize : tuple
        A tuple that determines the figure size. If None, size is
        determined based on the number of variables in the model
    resid : bool
        If True, use residuals as the response variable
    """

    exog_names = extract_evars(fitted.model, df.columns)

    if resid:
        endog = fitted.resid
        endog_name = "residuals"
        if df.shape[0] != endog.shape[0]:
            raise ValueError(
                "The number of observations in the fitted model and the data frame must be the same"
            )
        else:
            df = pd.concat(
                [
                    pd.DataFrame({endog_name: endog}),
                    df[exog_names].copy().reset_index(drop=True),
                ],
                axis=1,
            )
    else:
        endog_name = extract_rvar(fitted.model, df.columns)
        df = df[[endog_name] + exog_names].copy()

    nr_plots = len(exog_names)
    if figsize is None:
        figsize = (10, 2 * max(nr_plots, 4))

    fig, ax = plt.subplots(max(ceil(nr_plots / 2), 2), 2, figsize=figsize)
    plt.subplots_adjust(wspace=0.25, hspace=0.25)

    idx = 0

    if nobs < df.shape[0] and nobs != np.inf and nobs != -1:
        df = df.copy().sample(nobs)

    while idx < nr_plots:
        row = idx // 2
        col = idx % 2
        exog_name = exog_names[idx]
        if pd.api.types.is_numeric_dtype(df[exog_name].dtype):
            fig = sns.scatterplot(x=exog_name, y=endog_name, data=df, ax=ax[row, col])
        else:
            fig = sns.stripplot(x=exog_name, y=endog_name, data=df, ax=ax[row, col])
            means = df.groupby(exog_name, observed=False)[endog_name].mean()
            levels = list(means.index)
            # Loop over categories
            for pos, cat in enumerate(levels):
                # Add a line for the mean
                ax[row, col].plot(
                    [pos - 0.5, pos + 0.5],
                    [means.iloc[pos], means.iloc[pos]],
                    color="blue",
                )

        idx += 1

    if nr_plots < 3:
        ax[-1, -1].remove()
        ax[-1, 0].remove()
        if nr_plots == 1:
            ax[0, -1].remove()
    elif nr_plots % 2 == 1:
        ax[-1, -1].remove()


def residual_plot(
    fitted,
    df,
    nobs: int = 1000,
    figsize: tuple = None,
) -> None:
    """
    Plot of variables vs residuals

    Parameters
    ----------
    fitted : A fitted linear regression model
    nobs : int
        Number of observations to use for the scatter plots. The default
        value is 1,000. To use all observations in the plots, use nobs=-1
    figsize : tuple
        A tuple that determines the figure size. If None, size is
        determined based on the number of variables in the model
    """

    scatter_plot(fitted, df, nobs=nobs, figsize=figsize, resid=True)


def cross_validation(mlobj, name, param_grid, scoring, directory="cv_objects", n_splits=5, n_jobs=2, random_state=1234):
    """
    Use cross-validation to find the best hyper parameters for a machine learning model

    Parameters
    ----------
    fitted : A fitted linear regression model
    nobs : int
        Number of observations to use for the scatter plots. The default
        value is 1,000. To use all observations in the plots, use nobs=-1
    figsize : tuple
        A tuple that determines the figure size. If None, size is
        determined based on the number of variables in the modelkkk
    mlobj: An object created by the pyrsm package with a fitted machine learning model
    name: str
        Name of the model
    param_grid: dict
        Dictionary with parameters to use for cross-validation
    scoring: dict
        Dictionary with scoring metrics to use for cross-validation
    directory : str
        Directory to save the cross-validation object. Defaults to "cv_objects"
    n_splits : int
        Number of splits to use in cross validation. Defaults to 5.
    n_jobs : int
        Number of just to start to run the cross validation. Defaults to 2.
    random_state : int:
        Random seed to use. Defaults to 1234.

    Returns:
        cv: An sklearn GridSearchCV object
    """

    cv_file = f"{directory}/{name}-cross-validation-object.pkl"
    if os.path.exists(cv_file):
        cv = load_state(cv_file)["cv"]
    else:
        rvar = mlobj.rvar
        if mlobj.mod_type == "classification":
            kfolds = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=random_state)
        else:
            kfolds = KFold(n_splits=n_splits, shuffle=True, random_state=random_state)
        cv = GridSearchCV(
            mlobj.fitted,
            param_grid=param_grid,
            scoring=scoring,
            cv=kfolds,
            n_jobs=n_jobs,
            refit=list(scoring.keys())[0],
            verbose=5,
        ).fit(mlobj.data_onehot, mlobj.data[rvar])
        if not os.path.exists(directory):
            os.mkdir(directory)
        save_state({"cv": cv}, cv_file)

    return cv



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from pathlib import Path
from shiny import App
from pyrsm.radiant.model.logistic import model_logistic
import pyrsm.radiant.utils as ru

www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
app_static = StaticFiles(directory=www_dir, html=False)

data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="titanic")
rc = model_logistic(data_dct, descriptions_dct, state=None, code=True)

routes = [
    Mount("/www", app=app_static),
    Mount("/", app=App(rc.shiny_ui, rc.shiny_server, debug=False)),
]
app = Starlette(debug=True, routes=routes)

# Prediction test for Data & Command option
# import pyrsm as rsm
# data_dct, descriptions_dct = rsm.radiant.utils.get_dfs(pkg="model", name="titanic")
# lr = rsm.model.logistic(
#   data=data_dct, rvar="price", evar=["carat", "clarity", "cut"]
# )
# reg.predict(data=diamonds, data_cmd={"carat": 1}, ci=True)
# reg.predict(data=diamonds, cmd={"carat": [1]}, ci=True)



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from pathlib import Path
from shiny import App
from pyrsm.radiant.model.xgboost import model_xgboost
import pyrsm.radiant.utils as ru

www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
app_static = StaticFiles(directory=www_dir, html=False)

data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="titanic")
rc = model_xgboost(data_dct, descriptions_dct, state=None, code=True)

routes = [
    Mount("/www", app=app_static),
    Mount("/", app=App(rc.shiny_ui, rc.shiny_server, debug=False)),
]
app = Starlette(debug=True, routes=routes)



from .central_limit_theorem import *
from .probability_calculator import *
from .single_mean import *
from .compare_means import *
from .single_prop import *
from .compare_props import *
from .cross_tabs import *
from .goodness import *
from .correlation import *



from shiny import App, Inputs, Outputs, Session, ui, reactive, render
import shinyswatch

# based on https://github.com/posit-dev/py-shinyswatch/issues/11#issuecomment-1647868499
from starlette.requests import Request as StarletteRequest

if "theme" not in globals():
    print("Define theme dictionary. Was not in globals()")
    theme = {"theme": "superhero"}


def app_ui(request: StarletteRequest):
    print("ui function was called")
    return ui.page_fluid(
        shinyswatch.get_theme(theme.get("theme", "superhero")),
        ui.tags.script(
            """
            Shiny.addCustomMessageHandler('refresh', function(message) {
                window.location.reload();
            });
            """
        ),
        ui.input_select(
            id="select_theme",
            label="Select a theme:",
            selected=theme.get("theme", "superhero"),
            choices=["superhero", "darkly", "sketchy"],
        ),
        # ui.output_ui("ui_select_theme"),
    )


def server(input: Inputs, output: Outputs, session: Session):
    print("server function was called")
    print(theme.get("theme", "superhero"))

    @reactive.Effect
    @reactive.event(input.select_theme, ignore_none=True)
    async def set_theme():
        if input.select_theme() != theme.get("theme", "superhero"):
            theme["theme"] = input.select_theme()
            await session.send_custom_message("refresh", "")
        # ui.update_select("select_theme", selected=theme.get("theme", "superhero"))

    # @output(id="ui_select_theme")
    # @render.ui
    # def ui_select_theme():
    #     return (
    #         ui.input_select(
    #             id="select_theme",
    #             label="Select a theme:",
    #             selected=theme.get("theme", "superhero"),
    #             choices=["superhero", "darkly", "sketchy"],
    #         ),
    #     )


app = App(app_ui, server, debug=False)



from typing import Union, Optional
import pandas as pd
import polars as pl
import numpy as np
from statsmodels.genmod.families import Binomial
from statsmodels.genmod.families.links import Logit
import statsmodels.formula.api as smf
from scipy import stats
from pyrsm.utils import ifelse, setdiff, check_dataframe
from pyrsm.model.visualize import distr_plot, pred_plot_sm, vimp_plot_sm, extract_evars, extract_rvar
from pyrsm.model.model import (
    sig_stars,
    model_fit,
    or_ci,
    or_plot,
    sim_prediction,
    predict_ci,
    check_binary,
    convert_to_list,
)
from pyrsm.model.model import vif as calc_vif
from pyrsm.basics.correlation import correlation


class logistic:
    """
    A class to perform logistic regression modeling (binary classification)

    Attributes
    ----------
    data : pd.DataFrame
        Dataset used for the analysis. If a Polars DataFrame is provided, it will be converted to a Pandas DataFrame.
    name : str
        Name of the dataset if provided as a dictionary.
    rvar : str
        Name of the response variable in the data.
    lev: str
        Name of the level in the response variable the model will predict.
    evar : list[str]
        List of column names of the explanatory variables.
    ivar : list[str]
        List of strings with the names of the columns included as explanatory variables (e.g., ["x1:x2", "x3:x4"])
    form : str
        Model specification formula.
    fitted : statsmodels.genmod.generalized_linear_model.GLMResultsWrapper
        The fitted model.
    coef : pd.DataFrame
        The estimated model coefficients with standard errors, p-values, etc.
    weights: pd.Series or None
        Frequency weights used in the model if provided.
    weights_name: str or None
        Column name for the variable in the data containing frequency weights if provided.

    Methods
    -------
    __init__(data, rvar=None, lev=None, evar=None, ivar=None, form=None)
        Initialize the logistic class with the provided data and parameters.
    summary(main=True, fit=True, ci=False, vif=False, test=None, dec=3)
        Summarize the model output.
    plot(plot_type, nobs=1000, incl=None, excl=None, incl_int=None, fix=True, hline=False, nnv=20, minq=0.025, maxq=0.975)
        Plot for the model.
    predict(data=None, cmd=None, data_cmd=None, ci=False, conf=0.95)
        Generate predictions using the fitted regression model.
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        rvar: Optional[str] = None,
        lev: Optional[str] = None,
        evar: Optional[list[str]] = None,
        ivar: Optional[list[str]] = None,
        form: Optional[str] = None,
        weights: Optional[str] = None,
    ) -> None:
        """
        Initialize the logistic class to build a logistic regression model with the provided data and parameters.

        Parameters
        ----------
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
            Dataset used for the analysis. If a Polars DataFrame is provided, it will be converted to a Pandas DataFrame. If a dictionary is provided, the key will be used as the name of the dataset.
        rvar : str, optional
            Name of the column in the data to be used as the response variable.
        lev: str
            Name of the level in the response variable the model will predict.
        evar : list[str], optional
            List of column names in the data to use as explanatory variables.
        ivar : list[str], optional
            List of interactions to add to the model as explanatory variables (e.g., ["x1:x2", "x3:x4])
        form : str, optional
            Optional formula to use if rvar and evar are not provided.
        """
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.rvar = rvar
        self.lev = lev
        self.evar = convert_to_list(evar)
        self.ivar = convert_to_list(ivar)
        self.form = form
        if weights is not None and weights != "None":
            self.weights_name = weights
            self.weights = self.data[weights]
        else:
            self.weights_name = self.weights = None

        if self.lev is not None and self.rvar is not None:
            self.data = check_binary(self.data, self.rvar, self.lev)

        if self.form:
            self.fitted = smf.glm(
                formula=self.form,
                data=self.data,
                freq_weights=self.weights,
                family=Binomial(link=Logit()),
            ).fit()
            self.evar = extract_evars(self.fitted.model, self.data.columns)
            self.rvar = extract_rvar(self.fitted.model, self.data.columns)
            if self.lev is None:
                self.lev = self.data.at[0, self.rvar]
        else:
            if self.evar is None or len(self.evar) == 0:
                self.form = f"{self.rvar} ~ 1"
            else:
                self.form = f"{self.rvar} ~ {' + '.join(self.evar)}"
            if self.ivar:
                self.form += f" + {' + '.join(self.ivar)}"
            self.fitted = smf.glm(
                formula=self.form,
                data=self.data,
                freq_weights=self.weights,
                family=Binomial(link=Logit()),
            ).fit()
        self.fitted.nobs_dropped = self.data.shape[0] - self.fitted.nobs
        df = pd.DataFrame(np.exp(self.fitted.params), columns=["OR"]).dropna()
        df["OR%"] = 100 * ifelse(df["OR"] < 1, -(1 - df["OR"]), df["OR"] - 1)
        df["coefficient"] = self.fitted.params
        df["std.error"] = self.fitted.params / self.fitted.tvalues
        # statsmodels uses "t.values" as the label
        df["z.value"] = self.fitted.tvalues
        df["p.value"] = self.fitted.pvalues
        df["  "] = sig_stars(self.fitted.pvalues)
        self.coef = df.reset_index()

    def summary(
        self, main=True, fit=True, ci=False, vif=False, test=None, dec=3
    ) -> None:
        """
        Summarize the logistic regression model output

        Parameters
        ----------
        main : bool, default True
            Print the main summary. Can be useful to turn off (i.e., False) when the focus is on other metrics (e.g., VIF).
        fit : bool, default True
            Print the fit statistics. Can be useful to turn off (i.e., False) when the focus is on other metrics (e.g., VIF).
        ci : bool, default False
            Print the confidence intervals for the coefficients.
        vif : bool, default False
            Print the generalized variance inflation factors.
        test : list[str] or None, optional
            List of variable names used in the model to test using a Chi-Square test or None if no tests are performed.
        dec : int, default 3
            Number of decimal places to round to.
        """
        if main:
            print("Logistic regression (GLM)")
            print(f"Data                 : {self.name}")
            print(f"Response variable    : {self.rvar}")
            print(f"Level                : {self.lev}")
            print(f"Explanatory variables: {', '.join(self.evar)}")
            if self.weights is not None:
                print(f"Weights used         : {self.weights_name}")
            print(f"Null hyp.: There is no effect of x on {self.rvar}")
            print(f"Alt. hyp.: There is an effect of x on {self.rvar}")

            df = self.coef.copy()
            df["OR"] = df["OR"].round(dec)
            df["coefficient"] = df["coefficient"].round(dec)
            df["std.error"] = df["std.error"].round(dec)
            df["z.value"] = df["z.value"].round(dec)
            df["p.value"] = ifelse(
                df["p.value"] < 0.001, "< .001", df["p.value"].round(dec)
            )
            df["OR%"] = [f"{round(o, max(dec-2, 0))}%" for o in df["OR%"]]
            df["index"] = df["index"].str.replace("[T.", "[", regex=False)
            df = df.set_index("index")
            df.index.name = None
            print(f"\n{df.to_string()}")
            print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

        if fit:
            print(f"\n{model_fit(self.fitted, dec=dec)}")

        if ci:
            print("\nConfidence intervals:")
            df = or_ci(self.fitted, dec=dec).set_index("index")
            df.index.name = None
            print(f"\n{df.to_string()}")

        if vif:
            if self.evar is None or len(self.evar) < 2:
                print("\nVariance Inflation Factors cannot be calculated")
            else:
                print("\nVariance inflation factors:")
                print(f"\n{calc_vif(self.fitted, dec=dec).to_string()}")

        if test is not None and len(test) > 0:
            self.chisq_test(test=test, dec=dec)

    def predict(
        self, data=None, cmd=None, data_cmd=None, ci=False, conf=0.95
    ) -> pd.DataFrame:
        """
        Generate probability predictions using the fitted model.

        Parameters
        ----------
        data : pd.DataFrame | pl.DataFrame, optional
            Data used to generate predictions. If not provided, the estimation data will be used.
        cmd : dict[str, Union[int, list[int]]], optional
            Dictionary with the names of the columns to be used in the prediction and the values to be used.
        data_cmd : dict[str, Union[int, list[int]]], optional
            Dictionary with the names of the columns to be used in the prediction and the values to be used.
        ci : bool, default False
            Calculate confidence intervals for the predictions.
        conf : float, default 0.95
            Confidence level for the intervals.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the predictions and the data used to make those predictions.

        """
        if data is None:
            data = self.data
        else:
            data = check_dataframe(data)
        data = data.loc[:, self.evar].copy()
        if data_cmd is not None:
            for k, v in data_cmd.items():
                data[k] = v
        elif cmd is not None:
            cmd = {k: ifelse(isinstance(v, str), [v], v) for k, v in cmd.items()}
            data = sim_prediction(data=data, vary=cmd)

        if ci:
            if data_cmd is not None:
                raise ValueError(
                    "Confidence intervals not available when using the Data & Command option"
                )
            else:
                return pd.concat(
                    [data, predict_ci(self.fitted, data, conf=conf)], axis=1
                )
        else:
            return data.assign(prediction=self.fitted.predict(data))

    def plot(
        self,
        plots="or",
        data=None,
        alpha=0.05,
        nobs: int = 1000,
        intercept=False,
        incl=None,
        excl=None,
        incl_int=[],
        fix=True,
        hline=False,
        nnv=20,
        minq=0.025,
        maxq=0.975,
        figsize=None,
        ax=None,
        ret=None,
    ) -> None:
        """
        Plots for a logistic regression model

        Parameters
        ----------
        plots : str or list[str], default 'dist'
            List of plot types to generate. Options include 'dist', 'corr', 'pred', 'vimp', 'or'.
        nobs : int, default 1000
            Number of observations to plot. Relevant for all plots that include a scatter of data points (i.e., corr, scatter, dashboard, residual).
        incl : list[str], optional
            Variables to include in the plot. Relevant for prediction plots (pred) and coefficient plots (coef).
        excl : list[str], optional
            Variables to exclude from the plot. Relevant for prediction plots (pred) and coefficient plots (coef).
        incl_int : list[str], optional
            Interaction terms to include in the plot. Relevant for prediction plots (pred).
        fix : bool, default True
            Fix the y-axis limits. Relevant for prediction plots (pred).
        hline : bool, default False
            Add a horizontal line to the plot at the mean of the response variable. Relevant for prediction plots (pred).
        nnv : int, default 20
            Number of predicted values to calculate and to plot. Relevant for prediction plots.
        minq : float, default 0.025
            Minimum quantile of the explanatory variable values to use to calculate and plot predictions.
        maxq : float, default 0.975
            Maximum quantile of the explanatory variable values to use to calculate and plot predictions.
        figsize : tuple[int, int], default None
            Figure size for the plots in inches.
        ax : plt.Axes, optional
            Axes object to plot on.
        ret : bool, optional
            Whether to return the variable (permutation) importance scores for a "vimp" plot.

        Examples
        --------
        >>> clf_lr.plot(plots='pdp')
        >>> clf_lr.plot(plots='pred', data=new_data)
        """
        plots = convert_to_list(plots)  # control for the case where a single string is passed
        excl = convert_to_list(excl)
        incl = ifelse(incl is None, None, convert_to_list(incl))
        incl_int = convert_to_list(incl_int)

        if data is None:
            data = self.data
        else:
            data = check_dataframe(data)
        if self.rvar in data.columns:
            data = data[[self.rvar] + self.evar].copy()
        else:
            data = data[self.evar].copy()
        if "dist" in plots:
            distr_plot(data),
        if "corr" in plots:
            cr = correlation(data)
            cr.plot(nobs=nobs, figsize=figsize)
        if "or" in plots or "coef" in plots:
            or_plot(
                self.fitted,
                alpha=alpha,
                intercept=intercept,
                incl=incl,
                excl=excl,
                figsize=figsize,
            )
        if "pred" in plots:
            pred_plot_sm(
                self.fitted,
                data=data,
                incl=incl,
                excl=excl,
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=nnv,
                minq=minq,
                maxq=maxq,
            )
        if "vimp" in plots or "pimp" in plots:
            return_vimp = vimp_plot_sm(
                self.fitted,
                data=data,
                rep=10,
                ax=ax,
                ret=ret,
            )
            if ret:
                return return_vimp

    def chisq_test(self, test=None, dec=3) -> None:
        """
        Chisq-test for competing models

        Parameters
        ----------
        test : list
            List of strings; contains the names of the columns of data to be tested
        dec : int, default 3
            Number of decimal places to round to.
        """
        if test is None:
            test = self.evar
        else:
            test = ifelse(isinstance(test, str), [test], test)

        evar = [c for c in self.evar if c not in test]
        if self.ivar is not None and len(self.ivar) > 0:
            sint = setdiff(self.ivar, test)
            test += [s for t in test for s in sint if f"I({t}" not in s and t in s]
            sint = setdiff(sint, test)
        else:
            sint = []

        form = f"{self.rvar} ~ "
        if len(evar) == 0 and len(sint) == 0:
            form += "1"
        else:
            form += f"{' + '.join(evar + sint)}"

        print(f"\nModel 1: {form}")
        print(f"Model 2: {self.form}")

        # LR test of competing models (slower than Wald but more accurate)
        sub_fitted = smf.glm(
            formula=form,
            data=self.data,
            freq_weights=self.weights,
            family=Binomial(link=Logit()),
        ).fit()

        lrtest = -2 * (sub_fitted.llf - self.fitted.llf)
        df = self.fitted.df_model - sub_fitted.df_model
        pvalue = stats.chi2.sf(lrtest, df)

        # calculate pseudo R-squared values for both models
        pr2_full = 1 - self.fitted.llf / self.fitted.llnull
        pr2_sub = 1 - sub_fitted.llf / sub_fitted.llnull

        print(f"Pseudo R-squared, Model 1 vs 2: {pr2_sub:.3f} vs {pr2_full:.3f}")
        pvalue = ifelse(pvalue < 0.001, "< .001", round(pvalue, dec))
        print(f"Chi-squared: {round(lrtest, dec)} df ({df:.0f}), p.value {pvalue}")



import matplotlib.pyplot as plt
import pandas as pd
import polars as pl
import numpy as np
from scipy import stats
from math import floor, ceil


def iround(x, dec):
    return x if not isinstance(x, float) else round(x, dec)


def check(lb, ub, plb, pub):
    if (lb is not None and ub is not None and lb > ub) or (
        plb is not None and pub is not None and plb > pub
    ):
        raise ValueError(
            "Please ensure the lower bound is smaller than the upper bound"
        )


def make_colors_discrete(ub, lb, x_range):
    if lb is not None and ub is not None:
        colors = [
            "red"
            if i < lb
            else "red"
            if i > ub
            else "green"
            if i == lb or i == ub
            else "blue"
            for i in x_range
        ]
    elif lb is not None:
        colors = ["red" if i < lb else "green" if i == lb else "blue" for i in x_range]
    elif ub is not None:
        colors = ["blue" if i < ub else "green" if i == ub else "red" for i in x_range]
    else:
        colors = "blue"
    return colors


def make_colors_continuous(ub, lb, x_range, y_range):
    fig, ax = plt.subplots()
    ax.plot(x_range, y_range, "k")
    if ub is not None and lb is not None:
        ax.fill_between(
            x_range,
            y_range,
            where=((x_range > lb) & (x_range < ub)),
            color="blue",
            alpha=0.5,
        )
        ax.fill_between(
            x_range, y_range, where=((x_range > ub) | (x_range < lb)), color="salmon"
        )
    elif ub is not None:
        ax.fill_between(
            x_range,
            y_range,
            where=(x_range < ub),
            color="blue",
            alpha=0.5,
        )
        ax.fill_between(x_range, y_range, where=(x_range > ub), color="salmon")
    elif lb is not None:
        ax.fill_between(
            x_range,
            y_range,
            where=(x_range > lb),
            color="blue",
            alpha=0.5,
        )
        ax.fill_between(x_range, y_range, where=(x_range < lb), color="salmon")
    else:
        ax.fill_between(
            x_range,
            y_range,
            where=((x_range > 0) | (x_range < 0)),
            color="blue",
            alpha=0.5,
        )

    if lb is not None:
        ax.axvline(lb, color="black", linestyle="dashed", linewidth=0.5)
    if ub is not None:
        ax.axvline(ub, color="black", linestyle="dashed", linewidth=0.5)

    return ax


def make_barplot(ub, lb, x_range, y_range):
    fig, ax = plt.subplots()
    colors = make_colors_discrete(ub, lb, x_range)
    ax.bar(x_range, y_range, color=colors, alpha=0.5)
    if len(x_range) <= 20:
        ax.set_xticks(x_range)

    return ax


def prob_binom(n: int, p: float, lb=None, ub=None, plb=None, pub=None):
    # Helper function to convert to int if not None
    def to_int(x):
        return int(max(0, x)) if x is not None else None

    lb = to_int(lb)
    ub = to_int(ub)

    (
        p_elb,
        p_lelb,
        p_lb,
        p_eub,
        p_leub,
        p_ub,
        p_int,
        vlb,
        vp_elb,
        vp_lelb,
        vp_lb,
        vub,
        vp_eub,
        vp_leub,
        vp_ub,
        vp_int,
    ) = [None] * 16

    if lb is not None:
        if lb > n:
            lb = n
        p_elb = stats.binom.pmf(lb, n, p)
        p_lelb = stats.binom.cdf(lb, n, p)
        if lb > 0:
            p_lb = sum(stats.binom.pmf(range(0, lb), n, p))
        else:
            p_lb = 0

    if ub is not None:
        if ub > n:
            ub = n
        p_eub = stats.binom.pmf(ub, n, p)
        p_leub = stats.binom.cdf(ub, n, p)
        if ub > 0:
            p_ub = sum(stats.binom.pmf(range(0, ub), n, p))
        else:
            p_ub = 0

    if lb is not None and ub is not None:
        p_int = sum(stats.binom.pmf(range(lb, ub + 1), n, p))

    if plb is not None:
        if plb > 1:
            plb = 1
        if plb < 0:
            plb = 0
        vlb = stats.binom.ppf(plb, n, p)
        vp_elb = stats.binom.pmf(vlb, n, p)
        vp_lelb = stats.binom.cdf(vlb, n, p)
        if vlb > 0:
            vp_lb = sum(stats.binom.pmf(range(0, int(vlb)), n, p))
        else:
            vp_lb = 0

    if pub is not None:
        if pub > 1:
            pub = 1
        if pub < 0:
            pub = 0
        vub = stats.binom.ppf(pub, n, p)
        vp_eub = stats.binom.pmf(vub, n, p)
        vp_leub = stats.binom.cdf(vub, n, p)
        if vub > 0:
            vp_ub = sum(stats.binom.pmf(range(0, int(vub)), n, p))
        else:
            vp_ub = 0

    check(lb, ub, plb, pub)

    if plb is not None and pub is not None:
        vp_int = sum(stats.binom.pmf(range(int(vlb), int(vub) + 1), n, p))

    # Return a dictionary of results
    return {
        "n": n,
        "p": p,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
        "p_elb": p_elb,
        "p_lb": p_lb,
        "p_eub": p_eub,
        "p_ub": p_ub,
        "p_int": p_int,
        "p_lelb": p_lelb,
        "p_leub": p_leub,
        "vlb": vlb,
        "vub": vub,
        "vp_elb": vp_elb,
        "vp_lb": vp_lb,
        "vp_eub": vp_eub,
        "vp_ub": vp_ub,
        "vp_lelb": vp_lelb,
        "vp_leub": vp_leub,
        "vp_int": vp_int,
    }


def summary_prob_binom(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: Binomial")

    n, p = dct["n"], dct["p"]
    print("n           :", n)
    print("p           :", p)
    print("Mean        :", round(n * p, dec))
    print("St. dev     :", round(np.sqrt(n * p * (1 - p)), dec))

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]
        p_elb, p_eub = dct["p_elb"], dct["p_eub"]
        p_lelb, p_leub = dct["p_lelb"], dct["p_leub"]

        print("Lower bound :", "" if lb is None else lb)
        print("Upper bound :", "" if ub is None else ub, "\n")

        if ub is not None or lb is not None:
            if lb is not None:
                # print(f"P(X  = {lb}) = {p_elb} (green)")
                print(f"P(X  = {lb}) = {p_elb}")
                if lb > 0:
                    print(f"P(X  < {lb}) = {p_lb}")
                    print(f"P(X <= {lb}) = {p_lelb}")
                if lb < n:
                    print(f"P(X  > {lb}) = {round(1 - (p_lb + p_elb), dec)}")
                    print(f"P(X >= {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                print(f"P(X  = {ub}) = {p_eub}")
                if ub > 0:
                    print(f"P(X  < {ub}) = {p_ub}")
                    print(f"P(X <= {ub}) = {p_leub}")
                if ub < n:
                    print(f"P(X  > {ub}) = {round(1 - (p_ub + p_eub), dec)}")
                    print(f"P(X >= {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} <= X <= {ub})     = {p_int}")
                print(f"1 - P({lb} <= X <= {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["vlb"], dct["vub"]
        vp_lb, vp_ub, vp_int = dct["vp_lb"], dct["vp_ub"], dct["vp_int"]
        vp_elb, vp_eub = dct["vp_elb"], dct["vp_eub"]
        vp_lelb, vp_leub = dct["vp_lelb"], dct["vp_leub"]

        print("Lower bound :", "" if plb is None else f"{plb} ({v_lb})")
        print("Upper bound :", "" if pub is None else f"{pub} ({v_ub})\n")

        if pub is not None or plb is not None:
            if plb is not None:
                print(f"P(X  = {v_lb}) = {vp_elb}")
                if v_lb > 0:
                    print(f"P(X  < {v_lb}) = {vp_lb}")
                    print(f"P(X <= {v_lb}) = {vp_lelb}")
                if v_lb < n:
                    print(f"P(X  > {v_lb}) = {round(1 - (vp_lb + vp_elb), dec)}")
                    print(f"P(X >= {v_lb}) = {round(1 - vp_lb, dec)}")

            if pub is not None:
                print(f"P(X  = {v_ub}) = {vp_eub}")
                if v_ub > 0:
                    print(f"P(X  < {v_ub}) = {vp_ub}")
                    print(f"P(X <= {v_ub}) = {vp_leub}")
                if v_ub < n:
                    print(f"P(X  > {v_ub}) = {round(1 - (vp_ub + vp_eub), dec)}")
                    print(f"P(X >= {v_ub}) = {round(1 - vp_ub, dec)}")

            if plb is not None and pub is not None:
                print(f"P({v_lb} <= X <= {v_ub})     = {vp_int}")
                print(f"1 - P({v_lb} <= X <= {v_ub}) = {round(1 - vp_int, dec)}")


def plot_prob_binom(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["vlb"], dct["vub"]

    n, p = dct["n"], dct["p"]

    x_range = np.arange(0, n)
    df = (
        pd.DataFrame()
        .assign(x_range=x_range, y_range=stats.binom.pmf(x_range, n, p))
        .query("y_range > 0.00001")
    )

    make_barplot(ub, lb, df.x_range, df.y_range)


def prob_chisq(df, lb=None, ub=None, plb=None, pub=None):
    if lb is not None:
        p_lb = stats.chi2.cdf(lb, df)
    else:
        p_lb = None

    if ub is not None:
        p_ub = stats.chi2.cdf(ub, df)
    else:
        p_ub = None

    if lb is not None and ub is not None:
        p_int = max(p_ub - p_lb, 0)
    else:
        p_int = None

    if plb is not None:
        v_lb = stats.chi2.ppf(plb, df)
    else:
        v_lb = None

    if pub is not None:
        v_ub = stats.chi2.ppf(pub, df)
    else:
        v_ub = None

    check(lb, ub, plb, pub)

    return {
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "v_lb": v_lb,
        "v_ub": v_ub,
        "df": df,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
    }


def summary_prob_chisq(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: Chi-square")

    df = dct["df"]
    print("Df          :", df)
    print("Mean        :", df)
    print("Variance    :", 2 * df)

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]

        print("Lower bound :", "0" if lb is None else lb)
        print("Upper bound :", "Inf" if ub is None else ub)

        if ub is not None or lb is not None:
            print()

            if lb is not None:
                print(f"P(X < {lb}) = {p_lb}")
                print(f"P(X > {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                print(f"P(X < {ub}) = {p_ub}")
                print(f"P(X > {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} < X < {ub})     = {p_int}")
                print(f"1 - P({lb} < X < {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["v_lb"], dct["v_ub"]

        print("Lower bound :", "0" if plb is None or plb < 0 else plb)
        print("Upper bound :", "1" if pub is None or pub > 1 else pub)

        if (pub is None or pub <= 1) or (plb is None or plb >= 0):
            print()

            if plb is not None and plb >= 0:
                print(f"P(X < {v_lb}) = {plb}")
                print(f"P(X > {v_lb}) = {round(1 - plb, dec)}")

            if pub is not None and pub <= 1:
                print(f"P(X < {v_ub}) = {pub}")
                print(f"P(X > {v_ub}) = {round(1 - pub, dec)}")

            if (plb is not None and plb >= 0) and (pub is not None and pub <= 1):
                print(f"P({v_lb} < X < {v_ub})     = {round(pub - plb, dec)}")
                print(f"1 - P({v_lb} < X < {v_ub}) = {round(1 - (pub - plb), dec)}")


def plot_prob_chisq(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["v_lb"], dct["v_ub"]

    df = dct["df"]
    x_range = np.linspace(
        floor(stats.chi2.ppf(0.001, df)), ceil(stats.chi2.ppf(1 - 0.001, df)), 1000
    )
    y_range = stats.chi2.pdf(x_range, df)
    make_colors_continuous(ub, lb, x_range, y_range)


def prob_disc(v, p, lb=None, ub=None, plb=None, pub=None):
    # Check lengths
    if len(v) != len(p):
        if len(v) % len(p) == 0:
            p *= len(v) // len(p)
        else:
            raise ValueError(
                "The number of values must be the same or a multiple of the number of probabilities"
            )

    # Ensure sum of probabilities is approximately 1
    if not 0.999 <= sum(p) <= 1.001:
        raise ValueError("Probabilities for a discrete variable do not sum to 1")

    # Sort values and probabilities in ascending order of values
    vp_sorted = sorted(zip(v, p))
    v, p = zip(*vp_sorted)

    # Discrete distribution functions
    def ddisc(b):
        return p[v.index(b)] if b in v else 0

    def pdisc(b):
        return sum(p[i] for i in range(len(v)) if v[i] < b)

    def qdisc(prob):
        return next((v[i] for i in range(len(p)) if sum(p[: i + 1]) >= prob), None)

    # Calculate probabilities
    result = {"v": v, "p": p}

    if lb is not None:
        result.update(
            {
                "lb": lb,
                "p_elb": ddisc(lb),
                "p_lb": pdisc(lb),
                "p_lelb": ddisc(lb) + pdisc(lb),
            }
        )
    if ub is not None:
        result.update(
            {
                "ub": ub,
                "p_eub": ddisc(ub),
                "p_ub": pdisc(ub),
                "p_leub": ddisc(ub) + pdisc(ub),
            }
        )
    if lb is not None and ub is not None:
        result["p_int"] = result["p_leub"] - result["p_lb"]

    if plb is not None:
        vlb = qdisc(plb)
        if vlb is not None:
            result.update(
                {
                    "plb": plb,
                    "vlb": vlb,
                    "vp_elb": ddisc(vlb),
                    "vp_lb": pdisc(vlb),
                    "vp_lelb": ddisc(vlb) + pdisc(vlb),
                }
            )
    if pub is not None:
        vub = qdisc(pub)
        if vub is not None:
            result.update(
                {
                    "pub": pub,
                    "vub": vub,
                    "vp_eub": ddisc(vub),
                    "vp_ub": pdisc(vub),
                    "vp_leub": ddisc(vub) + pdisc(vub),
                }
            )
    if (
        plb is not None
        and pub is not None
        and result.get("vlb") is not None
        and result.get("vub") is not None
    ):
        result["vp_int"] = result["vp_leub"] - result["vp_lb"]

    check(lb, ub, plb, pub)

    return result


def summary_prob_disc(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    lb, ub = dct.get("lb", None), dct.get("ub", None)
    v, p = dct["v"], dct["p"]
    mean = sum(vi * pi for vi, pi in zip(v, p))
    std_dev = (sum(pi * (vi - mean) ** 2 for vi, pi in zip(v, p))) ** 0.5

    header = {
        "Distribution :": "Discrete",
        "Values       :": v,
        "Probabilities:": [round(i, dec) for i in p],
        "Mean         :": round(mean, dec),
        "St. dev      :": round(std_dev, dec),
        "Lower bound  :": None,
        "Upper bound  :": None,
    }
    summary = {}

    if type == "values":
        if lb is not None:
            header.update({"Lower bound  :": lb})
            summary.update({f"P(X  = {lb}) =": round(dct.get("p_elb", 0), dec)})
            if lb > min(v):
                summary.update(
                    {
                        f"P(X  < {lb}) =": dct.get("p_lb", 0),
                        f"P(X <= {lb}) =": dct.get("p_lelb", 0),
                    }
                )
            if lb < max(v):
                summary.update(
                    {
                        f"P(X  > {lb}) =": round(
                            1 - (dct.get("p_lb", 0) + dct.get("p_elb", 0)), dec
                        ),
                        f"P(X >= {lb}) =": round(1 - dct.get("p_lb", 0), dec),
                    }
                )

        if ub is not None:
            header.update({"Upper bound  :": ub})
            summary.update({f"P(X  = {ub}) =": dct.get("p_eub", 0)})
            if ub > min(v):
                summary.update(
                    {
                        f"P(X  < {ub}) =": dct.get("p_ub", 0),
                        f"P(X <= {ub}) =": dct.get("p_leub", 0),
                    }
                )
            if ub < max(v):
                summary.update(
                    {
                        f"P(X  > {ub}) =": round(
                            1 - (dct.get("p_ub", 0) + dct.get("p_eub", 0)), dec
                        ),
                        f"P(X >= {ub}) =": round(1 - dct.get("p_ub", 0), dec),
                    }
                )

        if lb is not None and ub is not None:
            summary.update(
                {
                    f"P({lb} <= X <= {ub})     =": dct.get("p_int", 0),
                    f"1 - P({lb} <= X <= {ub}) =": round(1 - dct.get("p_int", 0), dec),
                }
            )
    else:
        plb, pub, vlb, vub = (
            dct.get("plb", None),
            dct.get("pub", None),
            dct.get("vlb", None),
            dct.get("vub", None),
        )

        if plb is not None:
            header.update({"Lower bound  :": f"{plb} ({dct.get('vlb')})"}),
            summary.update({f"P(X  = {vlb}) =": dct.get("vp_elb", 0)})
            if vlb > min(v):
                summary.update(
                    {
                        f"P(X  < {vlb}) = ": dct.get("vp_lb", 0),
                        f"P(X <= {vlb}) =": dct.get("vp_lelb", 0),
                    }
                )
            if vlb < max(v):
                summary.update(
                    {
                        f"P(X  > {vlb}) =": round(
                            1 - (dct.get("vp_lb", 0) + dct.get("vp_elb", 0)), dec
                        ),
                        f"P(X >= {vlb}) =": round(1 - dct.get("vp_lb", 0), dec),
                    }
                )

        if pub is not None:
            header.update({"Upper bound  :": f"{pub} ({dct.get('vub')})"})
            summary.update({f"P(X  = {vub}) =": dct.get("vp_eub", 0)})
            if vub > min(v):
                summary.update(
                    {
                        f"P(X  < {vub}) =": dct.get("vp_ub", 0),
                        f"P(X <= {vub}) =": dct.get("vp_leub", 0),
                    }
                )
            if vub < max(v):
                summary.update(
                    {
                        f"P(X  > {vub}) = ": round(
                            1 - (dct.get("vp_ub", 0) + dct.get("vp_eub", 0)), dec
                        ),
                        f"P(X >= {vub}) =": round(1 - dct.get("vp_ub", 0), dec),
                    }
                )

        if plb is not None and pub is not None:
            summary.update(
                {
                    f"P({vlb} <= X <= {vub})     =": dct.get("vp_int", 0),
                    f"1 - P({vlb} <= X <= {vub}) =": round(
                        1 - dct.get("vp_int", 0), dec
                    ),
                }
            )

    for k, v in header.items():
        if v is not None:
            print(f"{k} {v}")

    for k, v in summary.items():
        print(f"{k} {v}")


def plot_prob_disc(dct, type="values"):
    if type == "values":
        lb, ub = dct.get("lb", None), dct.get("ub", None)
    else:
        lb, ub = dct.get("vlb", None), dct.get("vub", None)
    x_range, y_range = dct["v"], dct["p"]

    fig, ax = plt.subplots()
    colors = make_colors_discrete(ub, lb, x_range)
    ax.bar(range(len(x_range)), y_range, tick_label=x_range, color=colors, alpha=0.5)


def prob_expo(rate, lb=None, ub=None, plb=None, pub=None):
    if lb is not None and lb < 0:
        lb = 0
    if ub is not None and ub < 0:
        ub = 0

    if lb is None:
        p_lb = None
    else:
        p_lb = stats.expon.cdf(lb, scale=1 / rate)

    if ub is None:
        p_ub = None
    else:
        p_ub = stats.expon.cdf(ub, scale=1 / rate)

    p_int = None
    if lb is not None and ub is not None:
        if lb > ub:
            raise ValueError(
                "Please ensure the lower bound is smaller than the upper bound value"
            )
        else:
            p_int = max(p_ub - p_lb, 0)

    if pub is not None:
        if pub < 0:
            pub = 0
        elif pub > 1:
            pub = 1

    if plb is not None:
        if plb < 0:
            plb = 0
        elif plb > 1:
            plb = 1

    if plb is not None:
        v_lb = stats.expon.ppf(plb, scale=1 / rate)
    else:
        v_lb = None

    if pub is not None:
        v_ub = stats.expon.ppf(pub, scale=1 / rate)
    else:
        v_ub = None

    check(lb, ub, plb, pub)

    return {
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "v_lb": v_lb,
        "v_ub": v_ub,
        "rate": rate,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
    }


def summary_prob_expo(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: Exponential")

    rate = dct["rate"]
    print("Rate        :", rate)
    print("Mean        :", round(1 / rate, dec))
    print("Variance    :", round(rate**-2, dec))

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]

        print("Lower bound :", "0" if lb is None else lb)
        print("Upper bound :", "Inf" if ub is None else ub)

        if ub is not None or lb is not None:
            print()

            if lb is not None:
                print(f"P(X < {lb}) = {p_lb}")
                print(f"P(X > {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                print(f"P(X < {ub}) = {p_ub}")
                print(f"P(X > {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} < X < {ub})     = {p_int}")
                print(f"1 - P({lb} < X < {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["v_lb"], dct["v_ub"]

        print("Lower bound :", "0" if plb is None or plb < 0 else plb)
        print("Upper bound :", "1" if pub is None or pub > 1 else pub)

        if (pub is None or pub <= 1) or (plb is None or plb >= 0):
            print()

            if plb is not None and plb >= 0:
                print(f"P(X < {v_lb}) = {plb}")
                print(f"P(X > {v_lb}) = {round(1 - plb, dec)}")

            if pub is not None and pub <= 1:
                print(f"P(X < {v_ub}) = {pub}")
                print(f"P(X > {v_ub}) = {round(1 - pub, dec)}")

            if (plb is not None and plb >= 0) and (pub is not None and pub <= 1):
                print(f"P({v_lb} < X < {v_ub})     = {pub - plb}")
                print(f"1 - P({v_lb} < X < {v_ub}) = {round(1 - (pub - plb), dec)}")


def plot_prob_expo(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["v_lb"], dct["v_ub"]

    rate = dct["rate"]
    x_range = np.linspace(0, stats.expon.ppf(0.99, scale=1 / rate), 1000)
    y_range = stats.expon.pdf(x_range, scale=1 / rate)
    make_colors_continuous(ub, lb, x_range, y_range)


def prob_fdist(df1, df2, lb=None, ub=None, plb=None, pub=None):
    if lb is not None and lb < 0:
        lb = 0
    if ub is not None and ub < 0:
        ub = 0

    if lb is None:
        p_lb = None
    else:
        p_lb = stats.f.cdf(lb, df1, df2)

    if ub is None:
        p_ub = None
    else:
        p_ub = stats.f.cdf(ub, df1, df2)

    p_int = None
    if lb is not None and ub is not None:
        p_int = max(p_ub - p_lb, 0)

    if pub is not None:
        if pub < 0:
            pub = 0
        elif pub > 1:
            pub = 1

    if plb is not None:
        if plb < 0:
            plb = 0
        elif plb > 1:
            plb = 1

    if plb is not None:
        v_lb = stats.f.ppf(plb, df1, df2)
    else:
        v_lb = None

    if pub is not None:
        v_ub = stats.f.ppf(pub, df1, df2)
    else:
        v_ub = None

    check(lb, ub, plb, pub)

    return {
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "v_lb": v_lb,
        "v_ub": v_ub,
        "df1": df1,
        "df2": df2,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
    }


def summary_prob_fdist(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: F")

    df1, df2 = dct["df1"], dct["df2"]
    print("Df 1        :", df1)
    print("Df 2        :", df2)
    m = round(df2 / (df2 - 2), dec) if df2 > 2 else "NA"
    variance = (
        round(
            (2 * df2**2 * (df1 + df2 - 2)) / (df1 * (df2 - 2) ** 2 * (df2 - 4)), dec
        )
        if df2 > 4
        else np.nan
    )
    print("Mean        :", m)
    print("Variance    :", variance)

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]

        print("Lower bound :", "0" if lb is None else lb)
        print("Upper bound :", "Inf" if ub is None else ub)

        if ub is not None or lb is not None:
            print()

            if lb is not None:
                lb = round(lb, dec)
                print(f"P(X < {lb}) = {p_lb}")
                print(f"P(X > {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                ub = round(ub, dec)
                print(f"P(X < {ub}) = {p_ub}")
                print(f"P(X > {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} < X < {ub})     = {p_int}")
                print(f"1 - P({lb} < X < {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["v_lb"], dct["v_ub"]

        print("Lower bound :", "0" if plb is None or plb < 0 else plb)
        print("Upper bound :", "1" if pub is None or pub > 1 else pub)

        if (pub is None or pub <= 1) or (plb is None or plb >= 0):
            print()

            if plb is not None and plb >= 0:
                v_lb = round(v_lb, dec)
                print(f"P(X < {v_lb}) = {plb}")
                print(f"P(X > {v_lb}) = {round(1 - plb, dec)}")

            if pub is not None and pub <= 1:
                v_ub = round(v_ub, dec)
                print(f"P(X < {v_ub}) = {pub}")
                print(f"P(X > {v_ub}) = {round(1 - pub, dec)}")

            if (plb is not None and plb >= 0) and (pub is not None and pub <= 1):
                print(f"P({v_lb} < X < {v_ub})     = {round(pub - plb, dec)}")
                print(f"1 - P({v_lb} < X < {v_ub}) = {round(1 - (pub - plb), dec)}")


def plot_prob_fdist(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["v_lb"], dct["v_ub"]

    df1, df2 = dct["df1"], dct["df2"]
    x_range = np.linspace(0, stats.f.ppf(0.99, df1, df2), 1000)
    y_range = stats.f.pdf(x_range, df1, df2)
    make_colors_continuous(ub, lb, x_range, y_range)


def prob_lnorm(meanlog, sdlog, lb=None, ub=None, plb=None, pub=None):
    if lb is None:
        lb = -np.inf
    if ub is None:
        ub = np.inf

    p_ub = stats.lognorm.cdf(ub, sdlog, scale=np.exp(meanlog))
    p_lb = stats.lognorm.cdf(lb, sdlog, scale=np.exp(meanlog))
    p_int = max(p_ub - p_lb, 0)

    if pub is not None:
        if pub > 1:
            pub = 1
        if pub < 0:
            pub = 0

    if plb is not None:
        if plb > 1:
            plb = 1
        if plb < 0:
            plb = 0

    v_ub = (
        stats.lognorm.ppf(pub, sdlog, scale=np.exp(meanlog))
        if pub is not None
        else None
    )
    v_lb = (
        stats.lognorm.ppf(plb, sdlog, scale=np.exp(meanlog))
        if plb is not None
        else None
    )

    check(lb, ub, plb, pub)

    return {
        "meanlog": meanlog,
        "sdlog": sdlog,
        "lb": lb,
        "ub": ub,
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "plb": plb,
        "pub": pub,
        "v_lb": v_lb,
        "v_ub": v_ub,
    }


def summary_prob_lnorm(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    meanlog, sdlog = dct["meanlog"], dct["sdlog"]
    lb, v_lb = dct["lb"], dct["v_lb"]
    ub, v_ub = dct["ub"], dct["v_ub"]
    p_lb, plb = dct["p_lb"], dct["plb"]
    p_ub, pub = dct["p_ub"], dct["pub"]
    p_int = dct["p_int"]
    print(
        f"Probability calculator\nDistribution: Log Normal\nMean log    : {round(meanlog, dec)}"
    )
    print(f"St. dev log : {round(sdlog, dec)}")

    if type == "values":
        print(f"Lower bound : {lb if lb is not None else '-Inf'}")
        print(f"Upper bound : {ub if ub is not None else 'Inf'}\n")
        if ub != np.inf and lb != -np.inf:
            print(f"P(X < {lb if lb is not None else '-Inf'}) = {p_lb}")
            print(f"P(X > {lb if lb is not None else '-Inf'}) = {round(1 - p_lb, dec)}")
            print(f"P(X < {ub if ub is not None else 'Inf'}) = {p_ub}")
            print(f"P(X > {ub if ub is not None else 'Inf'}) = {round(1 - p_ub, dec)}")
            print(
                f"P({lb if lb is not None else '-Inf'} < X < {ub if ub is not None else 'Inf'})     = {round(p_int, dec)}"
            )
            print(
                f"1 - P({lb if lb is not None else '-Inf'} < X < {ub if ub is not None else 'Inf'}) = {round(1 - p_int, dec)}"
            )
        elif lb != -np.inf:
            print(f"P(X < {lb if lb is not None else '-Inf'}) = {p_lb}")
            print(f"P(X > {lb if lb is not None else '-Inf'}) = {round(1 - p_lb, dec)}")
        elif ub != np.inf:
            print(f"P(X < {ub if ub is not None else 'Inf'}) = {p_ub}")
            print(f"P(X > {ub if ub is not None else 'Inf'}) = {round(1 - p_ub, dec)}")
    else:
        if pub is None:
            pub = 2
        if plb is None:
            plb = -1

        print("Lower bound :", "0" if plb < 0 else plb, "(" + str(v_lb) + ")")
        print("Upper bound :", "1" if pub > 1 else pub, "(" + str(v_ub) + ")")

        if pub <= 1 or plb >= 0:
            print()

            if plb >= 0:
                print(f"P(X < {v_lb}) = {plb}")
                print(f"P(X > {v_lb}) = {round(1 - plb, dec)}")

            if pub <= 1:
                print(f"P(X < {v_ub}) = {pub}")
                print(f"P(X > {v_ub}) = {round(1 - pub, dec)}")

            if pub <= 1 and plb >= 0:
                print(f"P({v_lb} < X < {v_ub})     = {round(pub - plb, dec)}")
                print(f"1 - P({v_lb} < X < {v_ub}) = {round(1 - (pub - plb), dec)}")


def plot_prob_lnorm(dct, type="values"):
    meanlog, sdlog = dct["meanlog"], dct["sdlog"]
    lb = dct["lb"] if type == "values" else dct["v_lb"]
    ub = dct["ub"] if type == "values" else dct["v_ub"]

    def scale(lb, ub):
        if (ub is None or abs(ub) == np.inf) and (lb is None or abs(lb) == np.inf):
            return 3
        elif ub is not None and abs(ub) != np.inf:
            return max(3, ub)
        elif lb is not None and abs(lb) != np.inf:
            return max(3, lb)

    x_range = np.linspace(0, (np.exp(meanlog) + scale(lb, ub)) * sdlog, 1000)
    y_range = stats.lognorm.pdf(x_range, sdlog, scale=np.exp(meanlog))
    make_colors_continuous(ub, lb, x_range, y_range)


def prob_norm(mean, stdev, lb=None, ub=None, plb=None, pub=None):
    if lb is not None:
        p_lb = stats.norm.cdf(lb, mean, stdev)
    else:
        p_lb = None

    if ub is not None:
        p_ub = stats.norm.cdf(ub, mean, stdev)
    else:
        p_ub = None

    if lb is not None and ub is not None:
        p_int = max(p_ub - p_lb, 0)
    else:
        p_int = None

    if plb is not None:
        v_lb = stats.norm.ppf(plb, mean, stdev)
    else:
        v_lb = None

    if pub is not None:
        v_ub = stats.norm.ppf(pub, mean, stdev)
    else:
        v_ub = None

    check(lb, ub, plb, pub)

    return {
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "v_lb": v_lb,
        "v_ub": v_ub,
        "mean": mean,
        "stdev": stdev,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
    }


def summary_prob_norm(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: Normal")

    mean, stdev = dct["mean"], dct["stdev"]
    print("Mean        :", round(mean, dec))
    print("St. dev     :", round(stdev, dec))

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]

        print("Lower bound :", "-Inf" if lb is None else lb)
        print("Upper bound :", "Inf" if ub is None else ub)

        if ub is not None or lb is not None:
            print()

            if lb is not None:
                lb = round(lb, dec)
                print(f"P(X < {lb}) = {p_lb}")
                print(f"P(X > {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                ub = round(ub, dec)
                print(f"P(X < {ub}) = {p_ub}")
                print(f"P(X > {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} < X < {ub})     = {p_int}")
                print(f"1 - P({lb} < X < {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["v_lb"], dct["v_ub"]

        print("Lower bound :", "0" if plb is None or plb < 0 else plb)
        print("Upper bound :", "1" if pub is None or pub > 1 else pub)

        if (pub is None or pub <= 1) or (plb is None or plb >= 0):
            print()

            if plb is not None and plb >= 0:
                v_lb = round(v_lb, dec)
                print(f"P(X < {v_lb}) = {plb}")
                print(f"P(X > {v_lb}) = {round(1 - plb, dec)}")

            if pub is not None and pub <= 1:
                v_ub = round(v_ub, dec)
                print(f"P(X < {v_ub}) = {pub}")
                print(f"P(X > {v_ub}) = {round(1 - pub, dec)}")

            if (plb is not None and plb >= 0) and (pub is not None and pub <= 1):
                print(f"P({v_lb} < X < {v_ub})     = {round(pub - plb, dec)}")
                print(f"1 - P({v_lb} < X < {v_ub}) = {round(1 - (pub - plb), dec)}")


def plot_prob_norm(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["v_lb"], dct["v_ub"]

    mean, stdev = dct["mean"], dct["stdev"]
    x_range = np.linspace(mean - 3 * stdev, mean + 3 * stdev, 1000)
    y_range = stats.norm.pdf(x_range, mean, stdev)
    make_colors_continuous(ub, lb, x_range, y_range)


def prob_pois(lamb, lb=None, ub=None, plb=None, pub=None):
    if lamb <= 0:
        raise ValueError("Lambda must be positive")

    if lb is not None and lb < 0:
        lb = 0
    if ub is not None and ub < 0:
        ub = 0

    if lb is None:
        p_lb = None
        p_elb = None
    else:
        p_elb = stats.poisson.pmf(lb, lamb)
        p_lelb = stats.poisson.cdf(lb, lamb)
        if lb > 0:
            p_lb = p_lelb - p_elb
        else:
            p_lb = 0

    if ub is None:
        p_ub = None
        p_eub = None
    else:
        p_eub = stats.poisson.pmf(ub, lamb)
        p_leub = stats.poisson.cdf(ub, lamb)
        if ub > 0:
            p_ub = p_leub - p_eub
        else:
            p_ub = 0

    p_int = None
    if lb is not None and ub is not None:
        p_int = max(np.sum([stats.poisson.pmf(k, lamb) for k in range(lb, ub + 1)]), 0)

    if pub is not None:
        if pub < 0:
            pub = 0
        elif pub > 1:
            pub = 1

    if plb is not None:
        if plb < 0:
            plb = 0
        elif plb > 1:
            plb = 1

    if plb is None:
        v_lb = None
    else:
        v_lb = stats.poisson.ppf(plb, lamb)

    if pub is None:
        v_ub = None
    else:
        v_ub = stats.poisson.ppf(pub, lamb)

    check(lb, ub, plb, pub)

    return {
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "v_lb": v_lb,
        "v_ub": v_ub,
        "lamb": lamb,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
        "p_elb": p_elb,
        "p_eub": p_eub,
    }


def summary_prob_pois(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: Poisson")

    lamb = dct["lamb"]
    print("Lambda      :", lamb)
    print("Mean        :", lamb)
    print("Variance    :", lamb)

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]
        p_elb, p_eub = dct["p_elb"], dct["p_eub"]

        print("Lower bound :", "" if lb is None else lb)
        print("Upper bound :", "" if ub is None else ub)

        if ub is not None or lb is not None:
            print()

            if lb is not None:
                print(f"P(X  = {lb}) = {p_elb}")
                if lb > 0:
                    print(f"P(X  < {lb}) = {p_lb}")
                    print(f"P(X <= {lb}) = {round(p_lb + p_elb, dec)}")
                print(f"P(X  > {lb}) = {round(1 - (p_lb + p_elb), dec)}")
                print(f"P(X >= {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                print(f"P(X  = {ub}) = {p_eub}")
                if ub > 0:
                    print(f"P(X  < {ub}) = {p_ub}")
                    print(f"P(X <= {ub}) = {round(p_ub + p_eub, dec)}")
                print(f"P(X  > {ub}) = {round(1 - (p_ub + p_eub), dec)}")
                print(f"P(X >= {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} <= X <= {ub})     = {p_int}")
                print(f"1 - P({lb} <= X <= {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["v_lb"], dct["v_ub"]

        print("Lower bound :", "" if plb is None else f"{plb} ({v_lb})")
        print("Upper bound :", "" if pub is None else f"{pub} ({v_ub})")

        if (pub is None or pub <= 1) or (plb is None or plb >= 0):
            print()

            if plb is not None and plb >= 0:
                print(f"P(X  < {v_lb}) = {plb}")
                print(f"P(X  > {v_lb}) = {round(1 - plb, dec)}")

            if pub is not None and pub <= 1:
                print(f"P(X  < {v_ub}) = {pub}")
                print(f"P(X  > {v_ub}) = {round(1 - pub, dec)}")


def plot_prob_pois(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["v_lb"], dct["v_ub"]

    lmb = dct["lamb"]
    x_range = range(int(stats.poisson.ppf(1 - 0.00001, lmb)))
    y_range = [stats.poisson.pmf(k, lmb) for k in x_range]

    df = (
        pd.DataFrame()
        .assign(x_range=x_range, y_range=y_range)
        .query("y_range > 0.00001")
    )

    make_barplot(ub, lb, df.x_range, df.y_range)


def prob_tdist(df, lb=None, ub=None, plb=None, pub=None):
    if lb is None:
        p_lb = None
    else:
        p_lb = stats.t.cdf(lb, df)

    if ub is None:
        p_ub = None
    else:
        p_ub = stats.t.cdf(ub, df)

    p_int = None
    if lb is not None and ub is not None:
        p_int = max(p_ub - p_lb, 0)

    if pub is not None:
        if pub < 0:
            pub = 0
        elif pub > 1:
            pub = 1

    if plb is not None:
        if plb < 0:
            plb = 0
        elif plb > 1:
            plb = 1

    if plb is not None:
        v_lb = stats.t.ppf(plb, df)
    else:
        v_lb = None

    if pub is not None:
        v_ub = stats.t.ppf(pub, df)
    else:
        v_ub = None

    check(lb, ub, plb, pub)

    return {
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "v_lb": v_lb,
        "v_ub": v_ub,
        "df": df,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
    }


def summary_prob_tdist(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: t")

    df = dct["df"]
    n = df + 1
    print("Df          :", df)
    print("Mean        :", 0)
    if n > 2:
        print("St. dev     :", round(n / (n - 2), dec))
    else:
        print("St. dev     :", "NA")

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]

        print("Lower bound :", "-Inf" if lb is None else lb)
        print("Upper bound :", "Inf" if ub is None else ub)

        if ub is not None or lb is not None:
            print()

            if lb is not None:
                print(f"P(X < {lb}) = {p_lb}")
                print(f"P(X > {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                ub = round(ub, dec)
                print(f"P(X < {ub}) = {p_ub}")
                print(f"P(X > {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} < X < {ub})     = {p_int}")
                print(f"1 - P({lb} < X < {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["v_lb"], dct["v_ub"]

        print("Lower bound :", "0" if plb is None or plb < 0 else plb)
        print("Upper bound :", "1" if pub is None or pub > 1 else pub)

        if (pub is None or pub <= 1) or (plb is None or plb >= 0):
            print()

            if plb is not None and plb >= 0:
                v_lb = round(v_lb, dec)
                print(f"P(X < {v_lb}) = {plb}")
                print(f"P(X > {v_lb}) = {round(1 - plb, dec)}")

            if pub is not None and pub <= 1:
                v_ub = round(v_ub, dec)
                print(f"P(X < {v_ub}) = {pub}")
                print(f"P(X > {v_ub}) = {round(1 - pub, dec)}")

            if (plb is not None and plb >= 0) and (pub is not None and pub <= 1):
                print(f"P({v_lb} < X < {v_ub})     = {round(pub - plb, dec)}")
                print(f"1 - P({v_lb} < X < {v_ub}) = {round(1 - (pub - plb), dec)}")


def plot_prob_tdist(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["v_lb"], dct["v_ub"]

    x_range = np.linspace(-3, 3, 1000)
    y_range = stats.t.pdf(x_range, dct["df"])
    make_colors_continuous(ub, lb, x_range, y_range)


def prob_unif(min, max, lb=None, ub=None, plb=None, pub=None):
    if min > max:
        raise ValueError("The maximum value must be larger than the minimum value")

    if lb is not None:
        p_lb = stats.uniform.cdf(lb, min, max - min)
    else:
        p_lb = None

    if ub is not None:
        p_ub = stats.uniform.cdf(ub, min, max - min)
    else:
        p_ub = None

    if lb is not None and ub is not None:
        p_int = np.max(p_ub - p_lb, 0)
    else:
        p_int = None

    if plb is not None:
        v_lb = stats.uniform.ppf(plb, min, max - min)
    else:
        v_lb = None

    if pub is not None:
        v_ub = stats.uniform.ppf(pub, min, max - min)
    else:
        v_ub = None

    mean = (max + min) / 2
    stdev = np.sqrt((max - min) ** 2 / 12)

    check(lb, ub, plb, pub)

    return {
        "p_lb": p_lb,
        "p_ub": p_ub,
        "p_int": p_int,
        "v_lb": v_lb,
        "v_ub": v_ub,
        "mean": mean,
        "stdev": stdev,
        "min": min,
        "max": max,
        "lb": lb,
        "ub": ub,
        "plb": plb,
        "pub": pub,
    }


def summary_prob_unif(dct, type="values", dec=3):
    dct = {k: iround(v, dec) for k, v in dct.items()}
    print("Probability calculator")
    print("Distribution: Uniform")

    min, max = dct["min"], dct["max"]
    mean, stdev = dct["mean"], dct["stdev"]

    print("Min         :", min)
    print("Max         :", max)
    print("Mean        :", round(mean, dec))
    print("St. dev     :", round(stdev, dec))

    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
        p_lb, p_ub, p_int = dct["p_lb"], dct["p_ub"], dct["p_int"]

        print("Lower bound :", min if lb is None else lb)
        print("Upper bound :", max if ub is None else ub)

        if ub is not None or lb is not None:
            print()

            if lb is not None:
                print(f"P(X < {lb}) = {p_lb}")
                print(f"P(X > {lb}) = {round(1 - p_lb, dec)}")

            if ub is not None:
                print(f"P(X < {ub}) = {p_ub}")
                print(f"P(X > {ub}) = {round(1 - p_ub, dec)}")

            if lb is not None and ub is not None:
                print(f"P({lb} < X < {ub})     = {p_int}")
                print(f"1 - P({lb} < X < {ub}) = {round(1 - p_int, dec)}")
    else:
        plb, pub = dct["plb"], dct["pub"]
        v_lb, v_ub = dct["v_lb"], dct["v_ub"]

        print("Lower bound :", "0" if plb is None or plb < 0 else plb)
        print("Upper bound :", "1" if pub is None or pub > 1 else pub)

        if (pub is None or pub <= 1) or (plb is None or plb >= 0):
            print()

            if plb is not None and plb >= 0:
                print(f"P(X < {v_lb}) = {plb}")
                print(f"P(X > {v_lb}) = {round(1 - plb, dec)}")

            if pub is not None and pub <= 1:
                print(f"P(X < {v_ub}) = {pub}")
                print(f"P(X > {v_ub}) = {round(1 - pub, dec)}")

            if (plb is not None and plb >= 0) and (pub is not None and pub <= 1):
                print(f"P({v_lb} < X < {v_ub})     = {round(pub - plb, dec)}")
                print(f"1 - P({v_lb} < X < {v_ub}) = {round(1 - (pub - plb), dec)}")


def plot_prob_unif(dct, type="values"):
    if type == "values":
        lb, ub = dct["lb"], dct["ub"]
    else:
        lb, ub = dct["v_lb"], dct["v_ub"]

    min, max = dct["min"], dct["max"]

    x_range = np.linspace(min, max, 1000)
    y_range = stats.uniform.pdf(x_range, min, max - min)
    ax = make_colors_continuous(ub, lb, x_range, y_range)
    ax.axvline(min, ymin=0.048, ymax=0.952, color="black", linewidth=1)
    ax.axvline(max, ymin=0.048, ymax=0.952, color="black", linewidth=1)



import json
# import pandas as pd
# from shiny import App, ui, render, reactive
import matplotlib.pyplot as plt
import io
from contextlib import redirect_stdout
import base64
from pyrsm.basics import single_mean, compare_means
from pyrsm.model import regress
import pyrsm as rsm
import pandas as pd

# data should only be loaded through the fileUpload feature
# added data here because otherwise we get an error that
# the relevant data is not available
base_dir = "pyrsm/data"
demand_uk = pd.read_parquet(f"{base_dir}/basics/demand_uk.parquet")
with open(f"{base_dir}/basics/demand_uk_description.md", "r") as f:
    demand_uk_description = f.read()

salary = pd.read_parquet(f"{base_dir}/basics/salary.parquet")
with open(f"{base_dir}/basics/salary_description.md", "r") as f:
    salary_description = f.read()


async def handle_single_mean(tool_call, chat):
    try:
        arguments = json.loads(tool_call.function.arguments)
        arguments["data"] = f"{{\"{arguments['data']}\": {arguments['data']}}}"

        cmd = (
            f"single_mean("
            + ", ".join(
                [
                    f"{k}={v}" if isinstance(v, (int, float)) or k == "data" else f'{k}="{v}"'
                    for k, v in arguments.items()
                    if k not in ["dec"]
                ]
            )
            + ")"
        )

        print(cmd)

        # Execute the analysis
        obj = eval(
            cmd,
        )

        # Extract additional arguments for summary method if provided
        dec = arguments.pop("dec", 3)

        # Get summary
        buffer_summary = io.StringIO()
        with redirect_stdout(buffer_summary):
            obj.summary(dec=dec)

        # Generate and send plot
        plt.figure()  # Create new figure
        obj.plot()
        buffer_plot = io.BytesIO()
        plt.savefig(buffer_plot, format="png")
        plt.close("all")  # Close all figures
        buffer_plot.seek(0)
        encoded_image = base64.b64encode(buffer_plot.read()).decode("utf-8")

        cmd = f"sm = {cmd}"
        cmd += f"\nsm.summary(dec={dec})"
        cmd += '\nsm.plot(plots="hist")'

        results = f"Generated code:\n```python\n{cmd}\n```\n"
        results += f"\nSummary output:\n```bash\n{buffer_summary.getvalue()}```\n"
        results += f"![plot](data:image/png;base64,{encoded_image})"
        await chat.append_message({"role": "assistant", "content": results})

    except Exception as e:
        error_msg = f"Error executing analysis: {str(e)}"
        await chat.append_message({"role": "assistant", "content": error_msg})
    finally:
        plt.close("all")  # Ensure all plots are closed


async def handle_compare_means(tool_call, chat):
    try:
        arguments = json.loads(tool_call.function.arguments)
        arguments["data"] = f"{{\"{arguments['data']}\": {arguments['data']}}}"

        cmd = (
            f"compare_means("
            + ", ".join(
                [
                    f"{k}={v}" if isinstance(v, (int, float)) or k == "data" else f'{k}="{v}"'
                    for k, v in arguments.items()
                    if k not in ["dec", "extra", "plots", "nobs"]
                ]
            )
            + ")"
        )

        # Execute the analysis
        obj = eval(cmd)

        # Extract additional arguments for summary method if provided
        dec = arguments.pop("dec", 3)
        extra = arguments.pop("extra", False)

        # Get summary
        buffer_summary = io.StringIO()
        with redirect_stdout(buffer_summary):
            obj.summary(extra=extra, dec=dec)

        # Extract plot type if provided in the arguments
        plots = arguments.pop("plots", "scatter")
        nobs = arguments.pop("nobs", None)

        # Generate and send plot
        plt.figure()  # Create new figure
        obj.plot(plots=plots, nobs=nobs)
        buffer = io.BytesIO()
        plt.savefig(buffer, format="png")
        plt.close("all")  # Close all figures
        buffer.seek(0)
        encoded_image = base64.b64encode(buffer.read()).decode("utf-8")

        cmd = f"cm = {cmd}"
        cmd += f"\ncm.summary(extra={extra}, dec={dec})"
        if plots in ["scatter", "box"]:
            cmd += f"\ncm.plot(plots=\"{plots}\", nobs={nobs})"
        else:
            cmd += f"\ncm.plot(plots=\"{plots}\")"

        results = f"Generated code:\n```python\n{cmd}\n```\n"
        results += f"\nSummary output:\n```bash\n{buffer_summary.getvalue()}```\n"
        results += f"![plot](data:image/png;base64,{encoded_image})"

        await chat.append_message({"role": "assistant", "content": results})

    except Exception as e:
        error_msg = f"Error executing analysis: {str(e)}"
        await chat.append_message({"role": "assistant", "content": error_msg})
    finally:
        plt.close("all")  # Ensure all plots are closed


async def handle_linear_regression(tool_call, chat):
    try:
        arguments = json.loads(tool_call.function.arguments)
        arguments["data"] = f"{{\"{arguments['data']}\": {arguments['data']}}}"
        # Handle comma, space, or plus separated variables
        separators = [" ", "+"]
        for sep in separators:
            arguments["evar"] = arguments["evar"].replace(sep, ",")
        arguments["evar"] = f"""{[x.strip() for x in arguments["evar"].split(",") if x.strip()]}"""

        # build a regression model with salary as the response variable and select several explanatory variables that can explain variation in salary.
        cmd = (
            f"regress("
            + ", ".join(
                [
                    f"{k}={v}" if isinstance(v, (int, float)) or k in ["data", "evar"] else f'{k}="{v}"'
                    for k, v in arguments.items()
                    if k not in ["dec", "extra", "plots", "nobs"]
                ]
            )
            + ")"
        )

        # Execute the analysis
        print(cmd)
        obj = eval(cmd)

        # Extract additional arguments for summary method if provided
        dec = arguments.pop("dec", 3)

        # Get summary
        buffer_summary = io.StringIO()
        with redirect_stdout(buffer_summary):
            obj.summary(dec=dec)

        # Extract plot type if provided in the arguments
        # plots = arguments.pop("plots", "scatter")
        # nobs = arguments.pop("nobs", None)

        # Generate and send plot
        # plt.figure()  # Create new figure
        # obj.plot(plots=plots, nobs=nobs)
        # buffer = io.BytesIO()
        # plt.savefig(buffer, format="png")
        # plt.close("all")  # Close all figures
        # buffer.seek(0)
        # encoded_image = base64.b64encode(buffer.read()).decode("utf-8")

        cmd = f"reg = {cmd}"
        cmd += f"\nreg.summary(dec={dec})"
        # if plots in ["scatter", "box"]:
        # cmd += f"\nreg.plot(plots=\"{plots}\", nobs={nobs})"
        # else:
        # cmd += f"\nreg.plot(plots=\"{plots}\")"

        results = f"Generated code:\n```python\n{cmd}\n```\n"
        results += f"\nSummary output:\n```bash\n{buffer_summary.getvalue()}```\n"
        # results += f"![plot](data:image/png;base64,{encoded_image})"

        await chat.append_message({"role": "assistant", "content": results})

    except Exception as e:
        error_msg = f"Error executing analysis: {str(e)}"
        await chat.append_message({"role": "assistant", "content": error_msg})
    finally:
        plt.close("all")  # Ensure all plots are closed



import streamlit as st
import pandas as pd
import statsmodels.formula.api as smf
import pyrsm as rsm
from datetime import datetime


def view_df(df):
    st.write("### Data:")
    st.write(df)


# @st.cache
def run_regression(df):
    st.write("### Select variables for linear Regression:")
    y = st.selectbox("Response variable:", df.columns)
    X = st.multiselect("Explanatory variables:", rsm.setdiff(df.columns, y))

    if X and y:
        form = f"{y} ~ {' + '.join(X)}"
        model = smf.ols(form, data=df).fit()

        st.write("### Summary of Linear regression (OLS):")
        st.write(model.summary())
        dt_string = datetime.now.strftime("%d/%m/%Y %H:%M:%S")
        st.write(f"Date and time: {dt_string}")


def sample_df(df):
    frac = st.slider("Random sample (%)", 1, 100, 100)

    if frac < 100:
        df = df.sample(frac=frac / 100, random_state=1234)

    return df


@st.cache
def get_df(file):
    extension = file.name.split(".")[-1].upper()
    if extension == "CSV":
        df = pd.read_csv(file)
    elif extension in ["XLS", "XLSX"]:
        df = pd.read_excel(file, engine="openpyxl")
    elif extension in ["PKL", "PICKLE"]:
        df = pd.read_pickle(file)
    return df


def main():
    st.write("## Regression model - Streamlit app")

    file = st.sidebar.file_uploader("Upload file", type=["csv", "xlsx", "pickle"])
    if not file:
        st.sidebar.write("Upload a .csv, .xlsx or .pickle file to get started")
        # need return and a 'main' function to avoid actions the
        # rest of the code is not ready for yet
        return None

    df = get_df(file)
    df = sample_df(df)

    st.sidebar.title("Visualization Selector")
    select_status = st.sidebar.radio("Function", ("View Data", "Regression(OLS)"))

    if select_status == "View Data":
        view_df(df)
    else:
        run_regression(df)


main()

# to run this app use the below from a terminal
# then open a browser at the url localhost:8501

# streamlit run streamlit-regression-app.py



from pyrsm.basics.compare_means import compare_means
from pyrsm.basics.correlation import correlation
from pyrsm.basics.cross_tabs import cross_tabs
import numpy as np
import pandas as pd
import polars as pl

df = pd.DataFrame({"x": [0, 1, 1, 1, 0, 0, 0], "y": [2, 1, 1, 1, 2, 2, 10]})
dfp = pl.DataFrame(df.copy())
dfp
df


# def test_compare_means():
#     cm = compare_means({"df": df}, var1="x", var2="y")
#     assert cm.
#     # assert c.cr[1, 0].round(3) == -0.493, "Correlations incorrect"
# df_nan = df.copy()
# df_nan.loc[4, "x"] = np.nan
# c = correlation(df_nan)
# assert c.cr[1, 0].round(3) == -0.567, "Correlations with np.nan incorrect"


def test_correlation_pandas():
    c = correlation(df)
    assert c.cr[1, 0].round(3) == -0.493, "Correlations incorrect"
    df_nan = df.copy()
    df_nan.loc[4, "x"] = np.nan
    c = correlation(df_nan)
    assert c.cr[1, 0].round(3) == -0.567, "Correlations with np.nan incorrect"


def test_correlation_polars():
    c = correlation(dfp)
    assert c.cr[1, 0].round(3) == -0.493, "Correlations incorrect"
    ## looks like an issue converting null to nan in pandas
    # df_nan = dfp
    # df_nan[4, "x"] = None
    # c = correlation(df_nan)
    # assert c.cr[1, 0].round(3) == -0.567, "Correlations with np.nan incorrect"


def test_crosstab():
    ct = cross_tabs(df, "x", "y")
    assert all(
        ct.expected.iloc[0, :].round(6) == [1.714286, 1.714286, 0.571429, 4.0000]
    ), "Cross tab expected values incorrect"


if __name__ == "__main__":
    test_correlation_pandas()
    test_correlation_polars()
    test_crosstab()



import io
import matplotlib.pyplot as plt
from shiny import render, ui, reactive, req
from contextlib import redirect_stdout, redirect_stderr
import pyrsm.radiant.utils as ru
from pyrsm.utils import intersect
import numpy as np
import pandas as pd
import pyrsm as rsm


def ui_predict(self, show_ci=True):
    if show_ci:
        ci = (
            ui.input_checkbox(
                "pred_ci",
                "Show conf. intervals",
                value=self.state.get("pred_ci", False),
            ),
            ui.panel_conditional(
                "input.pred_ci == true",
                ui.input_slider(
                    id="conf",
                    label="Confidence level:",
                    min=0,
                    max=1,
                    value=self.state.get("conf", 0.95),
                ),
            ),
        )
    else:
        ci = None

    return ui.panel_conditional(
        "input.tabs == 'Predict'",
        ui.panel_well(
            ui.input_select(
                "pred_type",
                "Prediction input type:",
                ["None", "Data", "Command", "Data & Command"],
                selected=self.state.get("pred_type", "None"),
            ),
            ui.panel_conditional(
                "input.pred_type == 'Data' || input.pred_type == 'Data & Command'",
                ui.input_select(
                    "pred_datasets",
                    "Prediction data:",
                    self.dataset_list,
                    selected=self.state.get("pred_data", None),
                ),
            ),
            ui.panel_conditional(
                "input.pred_type == 'Command' || input.pred_type == 'Data & Command'",
                ru.input_return_text_area(
                    "pred_cmd",
                    "Prediction command:",
                    rows=3,
                    value=self.state.get("pred_cmd", ""),
                    placeholder="Specify a dictionary of values to to use for prediction, e.g., {'carat': 1, 'cut': 'Ideal'}",
                ),
            ),
            ci,
        ),
    )


def make_model_inputs(self, input, output, get_data, mod_type="isBin"):
    if isinstance(mod_type, str):

        @output(id="ui_rvar")
        @render.ui
        def ui_rvar():
            isType = get_data()["var_types"][mod_type]
            return ui.input_select(
                id="rvar",
                label="Response Variable:",
                selected=self.state.get("rvar", None),
                choices=isType,
            )

    elif isinstance(mod_type, dict):

        @output(id="ui_rvar")
        @render.ui
        def ui_rvar():
            isType = get_data()["var_types"][mod_type[input.mod_type()]]
            return ui.input_select(
                id="rvar",
                label="Response Variable:",
                selected=self.state.get("rvar", None),
                choices=isType,
            )

    @output(id="ui_evar")
    @render.ui
    def ui_evar():
        vars = get_data()["var_types"]["all"].copy()
        if (input.rvar() is not None) and (input.rvar() in vars):
            del vars[input.rvar()]

        return ui.input_select(
            id="evar",
            label="Explanatory Variables:",
            selected=self.state.get("evar", []),
            choices=vars,
            multiple=True,
            size=min(8, len(vars)),
            selectize=False,
        )


def make_int_inputs(self, input, output, get_data):
    @output(id="ui_interactions")
    @render.ui
    def ui_interactions():
        if len(input.evar()) > 1 and input.show_interactions() is not None:
            choices = []
            nway = int(input.show_interactions())
            isNum = intersect(
                list(input.evar()), list(get_data()["var_types"]["isNum"].keys())
            )
            if len(isNum) > 0:
                # choices += ru.qterms(isNum, nway=int(input.show_interactions()[0]))
                choices += ru.qterms(isNum, nway=nway)
            choices += ru.iterms(input.evar(), nway=nway)
            return ui.input_select(
                id="interactions",
                label=None,
                selected=self.state.get("interactions", None),
                choices=choices,
                multiple=True,
                size=min(8, len(choices)),
                selectize=False,
            )

    @output(id="ui_evar_test")
    @render.ui
    def ui_evar_test():
        choices = {e: e for e in input.evar()}
        if (
            "show_interactions" in input
            and input.show_interactions() is not None
            and int(input.show_interactions()[0]) > 1
        ):
            choices.update({e: e for e in input.interactions()})
        return ui.input_selectize(
            id="evar_test",
            label="Variables to test:",
            selected=self.state.get("evar_test", None),
            choices=choices,
            multiple=True,
        )

    @output(id="ui_incl_evar")
    @render.ui
    def ui_incl_evar():
        if len(input.evar()) > 0:
            return ui.input_select(
                id="incl_evar",
                label="Explanatory variables to include:",
                selected=self.state.get("incl_evar", None),
                choices=input.evar(),
                multiple=True,
                size=min(8, len(input.evar())),
                selectize=False,
            )

    @output(id="ui_incl_interactions")
    @render.ui
    def ui_incl_interactions():
        if len(input.evar()) > 1:
            if "show_interactions" in input and input.show_interactions() is not None:
                nway = int(input.show_interactions())
            else:
                nway = 2
            choices = ru.iterms(input.evar(), nway=nway)
            return ui.input_select(
                id="incl_interactions",
                label="Interactions to include:",
                selected=self.state.get("incl_interactions", None),
                choices=choices,
                multiple=True,
                size=min(8, len(choices)),
                selectize=False,
            )


def make_estimate(
    self,
    input,
    output,
    get_data,
    fun,
    ret,
    module="model.",
    run=True,
    ec=None,
    debug=True,
):
    if ec is None:

        def estimation_code():
            data_name, code = (get_data()[k] for k in ["data_name", "code"])
            # from https://discord.com/channels/1109483223987277844/1127817202804985917/1129530385429176371
            # does not work with ifelse for some reason
            # inp = ifelse("lev" in input, input.lev(), None)
            if "lev" in input:
                inp = input.lev()
            else:
                inp = None
            if "weights" in input:
                weights = input.weights()
                if weights == "None":
                    weights = None
            else:
                weights = None

            args = {
                "data": f"""{{"{data_name}": {data_name}}}""",
                "rvar": input.rvar(),
                "lev": inp,
                "evar": list(input.evar()),
                "weights": weights,
            }

            if (
                "show_interactions" in input
                and input.show_interactions() is not None
                and int(input.show_interactions()) > 0
                and len(input.interactions()) > 0
            ):
                args["ivar"] = list(input.interactions())

            if ret == "xgb":
                args.update(
                    {
                        "mod_type": input.mod_type(),
                        "n_estimators": input.n_estimators(),
                        "max_depth": input.max_depth(),
                        "min_child_weight": input.min_child_weight(),
                        "learning_rate": input.learning_rate(),
                        "subsample": input.subsample(),
                        "colsample_bytree": input.colsample_bytree(),
                        "random_state": input.random_state(),
                    }
                )
            elif ret == "rf":
                if input.max_features() in ["sqrt", "log2"]:
                    max_features = input.max_features()
                else:
                    max_features = int(input.max_features())

                args.update(
                    {
                        "mod_type": input.mod_type(),
                        "n_estimators": input.n_estimators(),
                        "max_features": max_features,
                        "min_samples_leaf": input.min_samples_leaf(),
                        "max_samples": input.max_samples(),
                        "random_state": input.random_state(),
                    }
                )
            elif ret == "nn":
                args.update(
                    {
                        "mod_type": input.mod_type(),
                        "hidden_layer_sizes": eval(input.hidden_layer_sizes()),
                        "activation": input.activation(),
                        "solver": input.solver(),
                        "alpha": input.alpha(),
                        # "batch_size": input.batch_size(),
                        "learning_rate_init": input.learning_rate_init(),
                        "max_iter": input.max_iter(),
                    }
                )

            args_str = ru.drop_default_args(args, getattr(rsm, fun))

            if "extra_args" in input and input.extra_args() != "":
                args_str += ", " + input.extra_args()
            return f"""rsm.{module}{fun}({args_str})""", code

    else:
        estimation_code = ec

    if debug:

        @output(id="show_estimation_code")
        @render.text
        def show_estimation_code():
            out = io.StringIO()
            with redirect_stdout(out), redirect_stderr(out):
                eval(estimation_code()[0])

            return out.getvalue()

    def show_code():
        sc = estimation_code()
        return f"""{sc[1]}\n{ret} = {sc[0]}"""

    if run:

        @reactive.Calc
        @reactive.event(input.run, ignore_none=True)
        def estimate():
            locals()[input.datasets()] = self.datasets[
                input.datasets()
            ]  # get data into local scope

            ec = estimation_code()
            # exec for multi-line
            # https://stackoverflow.com/a/12698067/1974918
            exec(ec[1])
            return eval(ec[0])

    else:

        @reactive.Calc
        def estimate():
            locals()[input.datasets()] = self.datasets[
                input.datasets()
            ]  # get data into local scope

            ec = estimation_code()
            exec(ec[1])
            return eval(ec[0])

    return show_code, estimate


def make_summary(
    self, input, output, session, show_code, estimate, ret, sum_fun, sc=None
):
    if sc is None:

        def summary_code():
            if "controls" in input:
                args = {c: True for c in input.controls()}
            else:
                args = {}
            if (
                "evar_test" in input
                and input.evar_test() is not None
                and len(input.evar_test()) > 0
            ):
                args["test"] = list(input.evar_test())

            args_string = ru.drop_default_args(args, sum_fun)
            return f"""{ret}.summary({args_string})"""

    else:
        summary_code = sc

    @output(id="show_summary_code")
    @render.text
    def show_summary_code():
        cmd = f"""{show_code()}\n{summary_code()}"""
        return ru.code_formatter(cmd, self, input, session, id="copy_summary")

    @output(id="summary")
    @render.text
    def summary():
        out = io.StringIO()
        with redirect_stdout(out), redirect_stderr(out):
            locals()[ret] = estimate()  # get model object into local scope
            eval(summary_code())
        return out.getvalue()


def make_predict(
    self, input, output, session, show_code, estimate, ret, pred_fun, show_ci=False
):
    def predict_code():
        args = {}
        cmd = input.pred_cmd().strip()

        if input.pred_type() == "Data":
            args["data"] = input.pred_datasets()
        elif input.pred_type() == "Command" and not ru.is_empty(cmd):
            args["data"] = None
            args["cmd"] = cmd
        elif input.pred_type() == "Data & Command" and not ru.is_empty(cmd):
            args["data"] = input.pred_datasets()
            args["data_cmd"] = cmd

        if "pred_ci" in input:
            ci = input.pred_ci()
            if ci:
                args.update({"ci": ci, "conf": input.conf()})

        args_string = ru.drop_default_args(
            args, pred_fun, ignore=["data", "cmd", "data_cmd"]
        )
        return f"""{ret}.predict({args_string})"""

    @output(id="show_predict_code")
    @render.text
    def show_predict_code():
        return ru.code_formatter(
            f"""{show_code()}\npred = {predict_code()}""",
            self,
            input,
            session,
            id="copy_predict",
        )

    @output(id="predict")
    @render.data_frame
    def predict():
        if input.pred_type() != "None":
            locals()[ret] = estimate()  # get model object into local scope
            if input.pred_type() in ["Data", "Data & Command"]:
                locals()[input.pred_datasets()] = self.datasets[
                    input.pred_datasets()
                ]  # get prediction data into local scope
            pred = eval(predict_code())

            summary = "Viewing rows {start} through {end} of {total}"
            if pred.shape[0] > 100_000:
                pred = pred[:100_000]
                summary += " (100K rows shown)"

            return render.DataTable(pred.round(3), summary=summary)
        else:
            return None


def make_plot(self, input, output, session, show_code, estimate, ret, pc=None):
    if pc is None:

        def plot_code():
            plots = input.plots()
            if plots == "pred":
                incl = list(input.incl_evar())
                if "incl_interactions" in input:
                    incl_int = list(input.incl_interactions())
                else:
                    incl_int = []
                cmd = f""", incl={incl}"""
                if len(incl_int) > 0:
                    cmd += f""", incl_int={incl_int}"""
                cmd = f"""{ret}.plot(plots="{plots}" {cmd})"""
            elif plots == "corr":
                cmd = f"""{ret}.plot(plots="{plots}", nobs={input.nobs()})"""
            else:
                cmd = f"""{ret}.plot(plots="{plots}")"""
            return cmd

    else:
        plot_code = pc

    @output(id="show_plot_code")
    @render.text
    def show_plot_code():
        plots = input.plots()
        if plots != "None":
            cmd = f"""{show_code()}\n{plot_code()}"""
            return ru.code_formatter(cmd, self, input, session, id="copy_plot")

    @reactive.Calc
    def gen_plot():
        locals()[ret] = estimate()
        eval(f"""{plot_code()}""")
        fig = plt.gcf()
        width, height = fig.get_size_inches()  # Get the size in inches
        return fig, width * 96, height * 96

    @output(id="plot")
    @render.plot
    def plot():
        plots = input.plots()
        if plots != "None":
            return gen_plot()[0]

    @output(id="plot_container")
    @render.ui
    def plot_container():
        req(estimate(), input.plots())
        width, height = gen_plot()[1:]
        aspect_ratio = height / width
        # return ui.output_plot("plot", height=f"{height}px", width=f"{width}px")
        return ui.div(
            {"style": f"width: 100%; padding-bottom: {aspect_ratio * 100}%; position: relative;"},
            ui.div(
                {"style": "position: absolute; top: 0; left: 0; right: 0; bottom: 0;"},
                ui.output_plot("plot", height="100%", width="100%"),
            ),
        )



import os
import pandas as pd
from shiny import App, ui, render, reactive
import matplotlib.pyplot as plt
import traceback

import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "tools")))
import tools.tool_function_definitions as tfd
import tools.tool_handlers as th

from app_utils import load_dotenv
from openai import AsyncOpenAI
from pyrsm.basics import single_mean, compare_means
from pyrsm.model import regress
import pyrsm as rsm

tools = tfd.tools

load_dotenv()
llm = AsyncOpenAI(api_key=os.environ.get("OPENAI_API_KEY"))

# Load default data
base_dir = "pyrsm/data"
demand_uk = pd.read_parquet(f"{base_dir}/basics/demand_uk.parquet")
with open(f"{base_dir}/basics/demand_uk_description.md", "r") as f:
    demand_uk_description = f.read()

salary = pd.read_parquet(f"{base_dir}/basics/salary.parquet")
with open(f"{base_dir}/basics/salary_description.md", "r") as f:
    salary_description = f.read()

app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.h4("Data Management"),
        ui.input_file("data_file", "Upload Data (CSV/Parquet)", accept=[".csv", ".parquet"]),
        ui.input_file("desc_file", "Upload Description (MD/TXT)", accept=[".md", ".txt"]),
        ui.input_select("dataset_select", "Select Dataset", choices=["salary", "demand_uk"]),
        ui.output_ui("current_description"),
        width=400,
    ),
    ui.navset_tab(ui.nav_panel("Chat", ui.chat_ui("chat")), ui.nav_panel("Data", ui.output_ui("data"))),
    fillable_mobile=True,
)


def server(input):
    # Reactive values to store datasets and descriptions
    datasets = reactive.value({"demand_uk": demand_uk, "salary": salary})
    descriptions = reactive.value({"demand_uk": demand_uk_description, "salary": salary_description})
    available_datasets = reactive.value(["salary", "demand_uk"])

    @reactive.Effect
    @reactive.event(input.data_file)
    def _():
        file_info = input.data_file()
        if not file_info:
            return

        name = os.path.splitext(file_info["name"])[0]
        if file_info["name"].endswith(".csv"):
            df = pd.read_csv(file_info["datapath"])
        else:  # parquet
            df = pd.read_parquet(file_info["datapath"])

        current_datasets = datasets.get()
        current_datasets[name] = df
        datasets.set(current_datasets)

        current_available = available_datasets.get()
        if name not in current_available:
            current_available.append(name)
            available_datasets.set(current_available)

    @reactive.Effect
    @reactive.event(input.desc_file)
    def _():
        file_info = input.desc_file()
        if not file_info:
            return

        name = os.path.splitext(file_info["name"])[0]
        with open(file_info["datapath"], "r") as f:
            desc = f.read()

        current_descriptions = descriptions.get()
        current_descriptions[name] = desc
        descriptions.set(current_descriptions)

    @render.ui
    def current_description():
        selected = input.dataset_select()
        desc_dict = descriptions.get()
        return ui.markdown(desc_dict.get(selected, "No description available"))

    @render.ui
    @reactive.event(input.dataset_select)
    def data():
        selected = input.dataset_select()
        data_dict = datasets.get()
        if selected in data_dict:
            return ui.output_data_frame("selected_data")
        return ui.p("No data selected")

    @render.data_frame
    def selected_data():
        selected = input.dataset_select()
        data_dict = datasets.get()
        return data_dict.get(selected)

    chat = ui.Chat(
        id="chat",
        messages=[
            {"content": "Hello! How can I help you today?", "role": "assistant"},
        ],
    )

    @chat.on_user_submit
    async def _():
        try:
            # Get current dataset and description
            selected = input.dataset_select()
            desc_dict = descriptions.get()
            current_desc = desc_dict.get(selected, "No description available")

            await chat.append_message({"role": "system", "content": f"Available data: {selected} - {current_desc}"})
            messages = chat.messages(format="openai")

            response = await llm.chat.completions.create(
                model="gpt-4o",  # Fixed model name
                messages=messages,
                tools=tools,
                tool_choice="auto",
            )

            message = response.choices[0].message

            # Handle tool calls if present
            tool_calls = message.tool_calls
            if tool_calls:
                for tool_call in tool_calls:
                    if tool_call.function.name == "single_mean":
                        await th.handle_single_mean(tool_call, chat)
                    elif tool_call.function.name == "compare_means":
                        await th.handle_compare_means(tool_call, chat)
                    elif tool_call.function.name == "regress":
                        await th.handle_linear_regression(tool_call, chat)

                    # Get interpretation after tool execution
                    # await chat.append_message(
                    #     {
                    #         "role": "system",
                    #         "content": f"Provide a detailed interpretation of the results from calling the {tool_call} function. If a plot was created, please provide an interpretation of the plot as well.",
                    #     }
                    # )
                    # interpret_response = await llm.chat.completions.create(
                    #     model="gpt-4o", messages=chat.messages(format="openai")
                    # )
                    # # Append the interpretation
                    # await chat.append_message(
                    #     {"role": "assistant", "content": interpret_response.choices[0].message.content}
                    # )
            else:
                # If no tool calls, just append the response content
                await chat.append_message({"role": "assistant", "content": message.content})

        except Exception as e:
            error_msg = f"An error occurred: {str(e)}\n{traceback.format_exc()}"
            await chat.append_message({"role": "assistant", "content": error_msg})
        finally:
            plt.close("all")  # Ensure all plots are closed

    # return [ui.update_select("dataset_select", choices=available_datasets)]


app = App(app_ui, server)



from pathlib import Path
from shiny import App, ui
from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from pyrsm.radiant import utils as ru
from pyrsm.radiant.data.data_view import data_view
from pyrsm.radiant.basics.probability_calculator import basics_probability_calculator
from pyrsm.radiant.basics.single_mean import basics_single_mean
from pyrsm.radiant.basics.single_prop import basics_single_prop
from pyrsm.radiant.basics.compare_means import basics_compare_means
from pyrsm.radiant.basics.compare_props import basics_compare_props
from pyrsm.radiant.basics.cross_tabs import basics_cross_tabs
from pyrsm.radiant.basics.goodness import basics_goodness
from pyrsm.radiant.basics.correlation import basics_correlation
from pyrsm.radiant.model.regress import model_regress
from pyrsm.radiant.model.logistic import model_logistic
from pyrsm.radiant.model.mlp import model_mlp
from pyrsm.radiant.model.rforest import model_rforest

# import polars as pl
# import pandas as pd

# linear regression app
# polars reads dates from parquet just fine, pandas does not
# date format is maintained when using to_pandas
# data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="diamonds", polars=True)
# data_dct["diamonds"] = data_dct["diamonds"].to_pandas()
data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="diamonds")
data_dct.update({"diamonds100": data_dct["diamonds"].sample(100)})

# main path to www folder
www_dir = Path(__file__).parent.parent / "radiant" / "www"

# serve static files
app_static = StaticFiles(directory=www_dir, html=False)

# data > view
data_dct, descriptions_dct = ru.get_dfs(pkg="data")
rc = data_view(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_data = App(rc.shiny_ui, rc.shiny_server, debug=False)

# probability calculator mean app
rc = basics_probability_calculator(state=None, code=True, navbar=ru.radiant_navbar())
app_pc = App(rc.shiny_ui, rc.shiny_server, debug=False)

# single mean app
data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="demand_uk")
rc = basics_single_mean(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_sm = App(rc.shiny_ui, rc.shiny_server, debug=False)

# compare means app
data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="salary")
rc = basics_compare_means(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_cm = App(rc.shiny_ui, rc.shiny_server, debug=False)

# single prop
data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="consider")
rc = basics_single_prop(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_sp = App(rc.shiny_ui, rc.shiny_server, debug=False)

# compare props
data_dct, descriptions_dct = ru.get_dfs(pkg="data", name="titanic")
rc = basics_compare_props(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_cp = App(rc.shiny_ui, rc.shiny_server, debug=False)

# cross-tabs app
data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
rc = basics_cross_tabs(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_ct = App(rc.shiny_ui, rc.shiny_server, debug=False)

# goodness app
data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
rc = basics_goodness(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_gf = App(rc.shiny_ui, rc.shiny_server, debug=False)

# correlation app
data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="salary")
rc = basics_correlation(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_cr = App(rc.shiny_ui, rc.shiny_server, debug=False)

# linear regression app
data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="diamonds")
data_dct.update({"diamonds100": data_dct["diamonds"].sample(100)})

descriptions_dct.update({"diamonds100": descriptions_dct["diamonds"]})
rc = model_regress(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_regress = App(rc.shiny_ui, rc.shiny_server)

# logistic regression app
data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="titanic")
rc = model_logistic(
    data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar()
)
app_logistic = App(rc.shiny_ui, rc.shiny_server)

# random forest app
data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="diamonds")
# data_dctc, descriptions_dctc = ru.get_dfs(pkg="model", name="titanic")
# data_dct.update({"titanic": data_dctc})
# descriptions_dct.update({"titanic": descriptions_dctc})

rc = model_rforest(data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar())
app_rforest = App(rc.shiny_ui, rc.shiny_server)

# # mlp app
rc = model_mlp(data_dct, descriptions_dct, state=None, code=True, navbar=ru.radiant_navbar())
app_mlp = App(rc.shiny_ui, rc.shiny_server)

# ui_nav = ui.page_navbar(
#     ru.head_content(),
#     ru.radiant_navbar(),
#     ru.ui_stop(),
#     title="Radiant for Python",
#     inverse=False,
#     id="navbar_id",
# )
# app_home = App(ui_nav, None)

# ---- combine apps ----
routes = [
    Mount("/basics/prob-calc/", app=app_pc),
    Mount("/basics/single-mean/", app=app_sm),
    Mount("/basics/compare-means/", app=app_cm),
    Mount("/basics/single-prop/", app=app_sp),
    Mount("/basics/compare-props/", app=app_cp),
    Mount("/basics/cross-tabs/", app=app_ct),
    Mount("/basics/goodness/", app=app_gf),
    Mount("/basics/correlation/", app=app_cr),
    Mount("/models/regress/", app=app_regress),
    Mount("/models/logistic/", app=app_logistic),
    Mount("/models/rforest/", app=app_rforest),
    Mount("/models/mlp/", app=app_mlp),
    Mount("/www/", app=app_static),
    Mount("/", app=app_data),  # MUST BE LAST!!!!
]

app = Starlette(debug=True, routes=routes)

# some combination of https://github.com/rstudio/py-shiny/issues/482
# and multiple inheritance might work to have a shared data object
# https://chat.openai.com/share/7dca0c72-bf2d-49c4-8ce7-f7dfb88b10b8
